from libc.stdio cimport FILE

cdef extern from "sundials/sundials_types.h":
    ctypedef float realtype
    ctypedef unsigned int booleantype
    ctypedef long sunindextype

cdef extern from "sundials/sundials_version.h":
    
    #Fill a string with SUNDIALS version information */
    int SUNDIALSGetVersion(char *version, int len)

    # Fills integers with the major, minor, and patch release version numbers and a string with the release label.
    int SUNDIALSGetVersionNumber(int *major, int *minor, int *patch,
                                 char *label, int len);
                                             
cdef extern from "sundials/sundials_nvector.h":
    cdef enum N_Vector_ID:
        SUNDIALS_NVEC_SERIAL,
        SUNDIALS_NVEC_PARALLEL,
        SUNDIALS_NVEC_OPENMP,
        SUNDIALS_NVEC_PTHREADS,
        SUNDIALS_NVEC_PARHYP,
        SUNDIALS_NVEC_PETSC,
        SUNDIALS_NVEC_CUDA,
        SUNDIALS_NVEC_RAJA,
        SUNDIALS_NVEC_OPENMPDEV,
        SUNDIALS_NVEC_TRILINOS,
        SUNDIALS_NVEC_MANYVECTOR,
        SUNDIALS_NVEC_MPIMANYVECTOR,
        SUNDIALS_NVEC_MPIPLUSX,
        SUNDIALS_NVEC_CUSTOM

    struct _generic_N_Vector_Ops:
        pass
    struct _generic_N_Vector:
        pass
    ctypedef _generic_N_Vector *N_Vector
    ctypedef _generic_N_Vector_Ops *N_Vector_Ops
    ctypedef N_Vector *N_Vector_S

    struct _generic_N_Vector_Ops:
        N_Vector_ID (*nvgetvectorid)(N_Vector)
        N_Vector    (*nvclone)(N_Vector)
        N_Vector    (*nvcloneempty)(N_Vector)
        void        (*nvdestroy)(N_Vector)
        void        (*nvspace)(N_Vector, sunindextype *, sunindextype *)
        realtype*   (*nvgetarraypointer)(N_Vector)
        void        (*nvsetarraypointer)(realtype *, N_Vector)
        void*       (*nvgetcommunicator)(N_Vector)
        sunindextype (*nvgetlength)(N_Vector)
        void        (*nvlinearsum)(realtype, N_Vector, realtype, N_Vector, N_Vector)
        void        (*nvconst)(realtype, N_Vector)
        void        (*nvprod)(N_Vector, N_Vector, N_Vector)
        void        (*nvdiv)(N_Vector, N_Vector, N_Vector)
        void        (*nvscale)(realtype, N_Vector, N_Vector)
        void        (*nvabs)(N_Vector, N_Vector)
        void        (*nvinv)(N_Vector, N_Vector)
        void        (*nvaddconst)(N_Vector, realtype, N_Vector)
        realtype    (*nvdotprod)(N_Vector, N_Vector)
        realtype    (*nvmaxnorm)(N_Vector)
        realtype    (*nvwrmsnorm)(N_Vector, N_Vector)
        realtype    (*nvwrmsnormmask)(N_Vector, N_Vector, N_Vector)
        realtype    (*nvmin)(N_Vector)
        realtype    (*nvwl2norm)(N_Vector, N_Vector)
        realtype    (*nvl1norm)(N_Vector)
        void        (*nvcompare)(realtype, N_Vector, N_Vector)
        booleantype (*nvinvtest)(N_Vector, N_Vector)
        booleantype (*nvconstrmask)(N_Vector, N_Vector, N_Vector)
        realtype    (*nvminquotient)(N_Vector, N_Vector)
        int (*nvlinearcombination)(int, realtype*, N_Vector*, N_Vector)
        int (*nvscaleaddmulti)(int, realtype*, N_Vector, N_Vector*, N_Vector*)
        int (*nvdotprodmulti)(int, N_Vector, N_Vector*, realtype*)
    
        int (*nvlinearsumvectorarray)(int, realtype, N_Vector*, realtype, 
                                      N_Vector*, N_Vector*)
        int (*nvscalevectorarray)(int, realtype*, N_Vector*, N_Vector*)
        int (*nvconstvectorarray)(int, realtype, N_Vector*)
        int (*nvwrmsnormvectorarray)(int, N_Vector*, N_Vector*, realtype*)
        int (*nvwrmsnormmaskvectorarray)(int, N_Vector*, N_Vector*, N_Vector,
                                         realtype*)
        int (*nvscaleaddmultivectorarray)(int, int, realtype*, N_Vector*, 
                                          N_Vector**, N_Vector**)
        int (*nvlinearcombinationvectorarray)(int, int, realtype*, N_Vector**,
                                              N_Vector*)
    
        realtype (*nvdotprodlocal)(N_Vector, N_Vector)
        realtype (*nvmaxnormlocal)(N_Vector)
        realtype (*nvminlocal)(N_Vector)
        realtype (*nvl1normlocal)(N_Vector)
        booleantype (*nvinvtestlocal)(N_Vector, N_Vector)
        booleantype (*nvconstrmasklocal)(N_Vector, N_Vector, N_Vector)
        realtype (*nvminquotientlocal)(N_Vector, N_Vector)
        realtype (*nvwsqrsumlocal)(N_Vector, N_Vector)
        realtype (*nvwsqrsummasklocal)(N_Vector, N_Vector, N_Vector)

    struct _generic_N_Vector:
        void *content
        N_Vector_Ops ops


    # * FUNCTIONS *
    N_Vector N_VNewEmpty()
    void N_VFreeEmpty(N_Vector v)
    int N_VCopyOps(N_Vector w, N_Vector v)

    N_Vector_ID N_VGetVectorID(N_Vector w)
    N_Vector N_VClone(N_Vector w)
    N_Vector N_VCloneEmpty(N_Vector w)
    void N_VDestroy(N_Vector v)
    void N_VSpace(N_Vector v, sunindextype *lrw, sunindextype *liw)
    realtype *N_VGetArrayPointer(N_Vector v)
    void N_VSetArrayPointer(realtype *v_data, N_Vector v)
    void *N_VGetCommunicator(N_Vector v)
    sunindextype N_VGetLength(N_Vector v)

    void N_VLinearSum(realtype a, N_Vector x, realtype b, N_Vector y, N_Vector z)
    void N_VConst(realtype c, N_Vector z)
    void N_VProd(N_Vector x, N_Vector y, N_Vector z)
    void N_VDiv(N_Vector x, N_Vector y, N_Vector z)
    void N_VScale(realtype c, N_Vector x, N_Vector z)
    void N_VAbs(N_Vector x, N_Vector z)
    void N_VInv(N_Vector x, N_Vector z)
    void N_VAddConst(N_Vector x, realtype b, N_Vector z)
    realtype N_VDotProd(N_Vector x, N_Vector y)
    realtype N_VMaxNorm(N_Vector x)
    realtype N_VWrmsNorm(N_Vector x, N_Vector w)
    realtype N_VWrmsNormMask(N_Vector x, N_Vector w, N_Vector id)
    realtype N_VMin(N_Vector x)
    realtype N_VWL2Norm(N_Vector x, N_Vector w)
    realtype N_VL1Norm(N_Vector x)
    void N_VCompare(realtype c, N_Vector x, N_Vector z)
    booleantype N_VInvTest(N_Vector x, N_Vector z)
    booleantype N_VConstrMask(N_Vector c, N_Vector x, N_Vector m)
    realtype N_VMinQuotient(N_Vector num, N_Vector denom)
    
    # /* OPTIONAL fused vector operations */
    int N_VLinearCombination(int nvec, realtype* c, N_Vector* X, N_Vector z)

    int N_VScaleAddMulti(int nvec, realtype* a, N_Vector x,
                         N_Vector* Y, N_Vector* Z)

    int N_VDotProdMulti(int nvec, N_Vector x, N_Vector* Y, realtype* dotprods)

    #/* OPTIONAL vector array operations */
    int N_VLinearSumVectorArray(int nvec, realtype a, N_Vector* X,
                                realtype b, N_Vector* Y, N_Vector* Z)

    int N_VScaleVectorArray(int nvec, realtype* c, N_Vector* X, N_Vector* Z)

    int N_VConstVectorArray(int nvec, realtype c, N_Vector* Z)

    int N_VWrmsNormVectorArray(int nvec, N_Vector* X, N_Vector* W, 
                               realtype* nrm)

    int N_VWrmsNormMaskVectorArray(int nvec, N_Vector* X, N_Vector* W, 
                                   N_Vector id, realtype* nrm)

    int N_VScaleAddMultiVectorArray(int nvec, int nsum, realtype* a, 
                                    N_Vector* X, N_Vector** Y, N_Vector** Z)

    int N_VLinearCombinationVectorArray(int nvec, int nsum, realtype* c, 
                                        N_Vector** X, N_Vector* Z)

    #/* OPTIONAL local reduction kernels (no parallel communication) */
    realtype N_VDotProdLocal(N_Vector x, N_Vector y)
    realtype N_VMaxNormLocal(N_Vector x)
    realtype N_VMinLocal(N_Vector x)
    realtype N_VL1NormLocal(N_Vector x)
    realtype N_VWSqrSumLocal(N_Vector x, N_Vector w)
    realtype N_VWSqrSumMaskLocal(N_Vector x, N_Vector w, N_Vector id)
    booleantype N_VInvTestLocal(N_Vector x, N_Vector z)
    booleantype N_VConstrMaskLocal(N_Vector c, N_Vector x, N_Vector m)
    realtype N_VMinQuotientLocal(N_Vector num, N_Vector denom)

    # * Additional functions exported by NVECTOR module

    N_Vector* N_VNewVectorArray(int count)
    N_Vector *N_VCloneEmptyVectorArray(int count, N_Vector w)
    N_Vector *N_VCloneVectorArray(int count, N_Vector w)
    void N_VDestroyVectorArray(N_Vector *vs, int count)

    #/* These function are really only for users of the Fortran interface */
    N_Vector N_VGetVecAtIndexVectorArray(N_Vector* vs, int index)
    void N_VSetVecAtIndexVectorArray(N_Vector* vs, int index, N_Vector w)

cdef extern from "sundials/sundials_matrix.h":
    cdef enum SUNMatrix_ID:
        SUNMATRIX_DENSE,
        SUNMATRIX_BAND,
        SUNMATRIX_SPARSE,
        SUNMATRIX_SLUNRLOC,
        SUNMATRIX_CUSTOM

    struct _generic_SUNMatrix_Ops:
        pass
    struct _generic_SUNMatrix:
        pass
    ctypedef _generic_SUNMatrix *SUNMatrix
    ctypedef _generic_SUNMatrix_Ops *SUNMatrix_Ops

    struct _generic_SUNMatrix_Ops:
        SUNMatrix_ID (*getid)(SUNMatrix)
        SUNMatrix    (*clone)(SUNMatrix)
        void         (*destroy)(SUNMatrix)
        int          (*zero)(SUNMatrix)
        int          (*copy)(SUNMatrix, SUNMatrix)
        int          (*scaleadd)(realtype, SUNMatrix, SUNMatrix)
        int          (*scaleaddi)(realtype, SUNMatrix)
        int          (*matvecsetup)(SUNMatrix);
        int          (*matvec)(SUNMatrix, N_Vector, N_Vector)
        int          (*space)(SUNMatrix, long int*, long int*)

    struct _generic_SUNMatrix:
        void *content
        SUNMatrix_Ops ops

    # * FUNCTIONS *
    SUNMatrix SUNMatNewEmpty()
    void SUNMatFreeEmpty(SUNMatrix A)
    int SUNMatCopyOps(SUNMatrix A, SUNMatrix B)
    SUNMatrix_ID SUNMatGetID(SUNMatrix A)
    SUNMatrix SUNMatClone(SUNMatrix A)
    void SUNMatDestroy(SUNMatrix A)
    int SUNMatZero(SUNMatrix A)
    int SUNMatCopy(SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAdd(realtype c, SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAddI(realtype c, SUNMatrix A)
    int SUNMatMatvecSetup(SUNMatrix A)
    int SUNMatMatvec(SUNMatrix A, N_Vector x, N_Vector y)
    int SUNMatSpace(SUNMatrix A, long int *lenrw, long int *leniw)

cdef extern from "sundials/sundials_iterative.h":
    enum:
        PREC_NONE
        PREC_LEFT
        PREC_RIGHT
        PREC_BOTH
    enum:
        MODIFIED_GS = 1
        CLASSICAL_GS = 2
    ctypedef int (*ATimesFn)(void *A_data, N_Vector v, N_Vector z)
    ctypedef int (*PSetupFn)(void *P_data)
    ctypedef int (*PSolveFn)(void *P_data, N_Vector r, N_Vector z,
                             realtype tol, int lr)

    int ModifiedGS(N_Vector *v, realtype **h, int k, int p,
                   realtype *new_vk_norm)
    int ClassicalGS(N_Vector *v, realtype **h, int k, int p,
                    realtype *new_vk_norm, N_Vector temp, realtype *s)
    int QRfact(int n, realtype **h, realtype *q, int job)
    int QRsol(int n, realtype **h, realtype *q, realtype *b)

    enum: SUNMAT_SUCCESS                  #    0  /* function successfull          */
    enum: SUNMAT_ILL_INPUT                # -701  /* illegal function input        */
    enum: SUNMAT_MEM_FAIL                 # -702  /* failed memory access/alloc    */
    enum: SUNMAT_OPERATION_FAIL           # -703  /* a SUNMatrix operation returned nonzero */
    enum: SUNMAT_MATVEC_SETUP_REQUIRED    # -704  /* the SUNMatMatvecSetup routine needs to be called */

cdef extern from "sundials/sundials_linearsolver.h":

    cdef enum SUNLinearSolver_Type:
        SUNLINEARSOLVER_DIRECT,
        SUNLINEARSOLVER_ITERATIVE,
        SUNLINEARSOLVER_MATRIX_ITERATIVE


    cdef enum SUNLinearSolver_ID:
        SUNLINEARSOLVER_BAND,
        SUNLINEARSOLVER_DENSE,
        SUNLINEARSOLVER_KLU,
        SUNLINEARSOLVER_LAPACKBAND,
        SUNLINEARSOLVER_LAPACKDENSE,
        SUNLINEARSOLVER_PCG,
        SUNLINEARSOLVER_SPBCGS,
        SUNLINEARSOLVER_SPFGMR,
        SUNLINEARSOLVER_SPGMR,
        SUNLINEARSOLVER_SPTFQMR,
        SUNLINEARSOLVER_SUPERLUDIST,
        SUNLINEARSOLVER_SUPERLUMT,
        SUNLINEARSOLVER_CUSOLVERSP_BATCHQR,
        SUNLINEARSOLVER_CUSTOM
        
    struct _generic_SUNLinearSolver_Ops:
        pass
    struct _generic_SUNLinearSolver:
        pass
    ctypedef _generic_SUNLinearSolver *SUNLinearSolver
    ctypedef _generic_SUNLinearSolver_Ops *SUNLinearSolver_Ops

    struct _generic_SUNLinearSolver_Ops:
        SUNLinearSolver_Type (*gettype)(SUNLinearSolver)
        SUNLinearSolver_ID   (*getid)(SUNLinearSolver);
        int                  (*setatimes)(SUNLinearSolver, void*, ATimesFn)
        int                  (*setpreconditioner)(SUNLinearSolver, void*,
                                                PSetupFn, PSolveFn)
        int                  (*setscalingvectors)(SUNLinearSolver,
                                                N_Vector, N_Vector)
        int                  (*initialize)(SUNLinearSolver)
        int                  (*setup)(SUNLinearSolver, SUNMatrix)
        int                  (*solve)(SUNLinearSolver, SUNMatrix, N_Vector,
                                    N_Vector, realtype)
        int                  (*numiters)(SUNLinearSolver)
        realtype             (*resnorm)(SUNLinearSolver)
        long int             (*lastflag)(SUNLinearSolver)
        int                  (*space)(SUNLinearSolver, long int*, long int*)
        N_Vector             (*resid)(SUNLinearSolver)
        int                  (*free)(SUNLinearSolver)

    struct _generic_SUNLinearSolver:
        void *content
        SUNLinearSolver_Ops ops

    SUNLinearSolver SUNLinSolNewEmpty()

    void SUNLinSolFreeEmpty(SUNLinearSolver S)

    SUNLinearSolver_Type SUNLinSolGetType(SUNLinearSolver S)

    SUNLinearSolver_ID SUNLinSolGetID(SUNLinearSolver S)
    int SUNLinSolSetATimes(SUNLinearSolver S, void* A_data,
                           ATimesFn ATimes)
    int SUNLinSolSetPreconditioner(SUNLinearSolver S, void* P_data,
                                   PSetupFn Pset, PSolveFn Psol)
    int SUNLinSolSetScalingVectors(SUNLinearSolver S, N_Vector s1,
                                   N_Vector s2)
    int SUNLinSolInitialize(SUNLinearSolver S)
    int SUNLinSolSetup(SUNLinearSolver S, SUNMatrix A)
    int SUNLinSolSolve(SUNLinearSolver S, SUNMatrix A, N_Vector x,
                       N_Vector b, realtype tol)
    int SUNLinSolNumIters(SUNLinearSolver S)
    realtype SUNLinSolResNorm(SUNLinearSolver S)
    N_Vector SUNLinSolResid(SUNLinearSolver S)
    long int SUNLinSolLastFlag(SUNLinearSolver S)
    int SUNLinSolSpace(SUNLinearSolver S, long int *lenrwLS,
                                       long int *leniwLS)
    int SUNLinSolFree(SUNLinearSolver S)

    enum: SUNLS_SUCCESS            #   0   /* successful/converged          */
    
    enum: SUNLS_MEM_NULL           # -801   /* mem argument is NULL          */
    enum: SUNLS_ILL_INPUT          # -802   /* illegal function input        */
    enum: SUNLS_MEM_FAIL           # -803   /* failed memory access          */
    enum: SUNLS_ATIMES_FAIL_UNREC  # -804   /* atimes unrecoverable failure  */
    enum: SUNLS_PSET_FAIL_UNREC    # -805   /* pset unrecoverable failure    */
    enum: SUNLS_PSOLVE_FAIL_UNREC  # -806   /* psolve unrecoverable failure  */
    enum: SUNLS_PACKAGE_FAIL_UNREC # -807   /* external package unrec. fail  */
    enum: SUNLS_GS_FAIL            # -808   /* Gram-Schmidt failure          */
    enum: SUNLS_QRSOL_FAIL         # -809   /* QRsol found singular R        */
    enum: SUNLS_VECTOROP_ERR       # -810   /* vector operation error        */

    enum: SUNLS_RES_REDUCED        # 801   /* nonconv. solve, resid reduced */
    enum: SUNLS_CONV_FAIL          # 802   /* nonconvergent solve           */
    enum: SUNLS_ATIMES_FAIL_REC    # 803   /* atimes failed recoverably     */
    enum: SUNLS_PSET_FAIL_REC      # 804   /* pset failed recoverably       */
    enum: SUNLS_PSOLVE_FAIL_REC    # 805   /* psolve failed recoverably     */
    enum: SUNLS_PACKAGE_FAIL_REC   # 806   /* external package recov. fail  */
    enum: SUNLS_QRFACT_FAIL        # 807   /* QRfact found singular matrix  */
    enum: SUNLS_LUFACT_FAIL        # 808   /* LUfact found singular matrix  */

cdef extern from "sundials/sundials_direct.h":
    enum: SUNDIALS_DENSE
    enum: SUNDIALS_BAND

    cdef struct _DlsMat:
        int type
        sunindextype M
        sunindextype N
        sunindextype ldim
        sunindextype mu
        sunindextype ml
        sunindextype s_mu
        realtype *data
        sunindextype ldata
        realtype **cols

    ctypedef _DlsMat *DlsMat

    DlsMat NewDenseMat(sunindextype M, sunindextype N)
    DlsMat NewBandMat(sunindextype N, sunindextype mu, sunindextype ml, sunindextype smu)
    void DestroyMat(DlsMat A)
    int *NewIntArray(int N)
    sunindextype *NewIndexArray(sunindextype N)
    realtype *NewRealArray(long int N)
    void DestroyArray(void *p)
    void AddIdentity(DlsMat A)
    void SetToZero(DlsMat A)
    void PrintMat(DlsMat A, FILE *outfile)

    # * Exported function prototypes (functions working on realtype**)
    realtype **newDenseMat(sunindextype m, sunindextype n)
    realtype **newBandMat(sunindextype n, sunindextype smu, sunindextype ml)
    void destroyMat(realtype **a)
    int *newIntArray(int n)
    sunindextype *newIndexArray(sunindextype n)
    realtype *newRealArray(sunindextype m)
    void destroyArray(void *v)

cdef extern from "sundials/sundials_band.h":
    sunindextype BandGBTRF(DlsMat A, sunindextype *p)
    sunindextype bandGBTRF(realtype **a, sunindextype n, sunindextype mu, sunindextype ml,
                       sunindextype smu, sunindextype *p)
    void BandGBTRS(DlsMat A, sunindextype *p, realtype *b)
    void bandGBTRS(realtype **a, sunindextype n, sunindextype smu, sunindextype ml,
                   sunindextype *p, realtype *b)
    void BandCopy(DlsMat A, DlsMat B, sunindextype copymu, sunindextype copyml)
    void bandCopy(realtype **a, realtype **b, sunindextype n, sunindextype a_smu,
                  sunindextype b_smu, sunindextype copymu, sunindextype copyml)
    void BandScale(realtype c, DlsMat A)
    void bandScale(realtype c, realtype **a, sunindextype n, sunindextype mu,
                   sunindextype ml, sunindextype smu)
    void bandAddIdentity(realtype **a, sunindextype n, sunindextype smu)
    void BandMatvec(DlsMat A, realtype *x, realtype *y)
    void bandMatvec(realtype **a, realtype *x, realtype *y, sunindextype n,
                    sunindextype mu, sunindextype ml, sunindextype smu)

cdef extern from "sundials/sundials_dense.h":
    sunindextype DenseGETRF(DlsMat A, sunindextype *p)
    void DenseGETRS(DlsMat A, sunindextype *p, realtype *b)

    sunindextype denseGETRF(realtype **a, sunindextype m, sunindextype n, sunindextype *p)
    void denseGETRS(realtype **a, sunindextype n, sunindextype *p, realtype *b)

    sunindextype DensePOTRF(DlsMat A)
    void DensePOTRS(DlsMat A, realtype *b)

    sunindextype densePOTRF(realtype **a, sunindextype m)
    void densePOTRS(realtype **a, sunindextype m, realtype *b)

    int DenseGEQRF(DlsMat A, realtype *beta, realtype *wrk)
    int DenseORMQR(DlsMat A, realtype *beta, realtype *vn, realtype *vm,
                   realtype *wrk)

    int denseGEQRF(realtype **a, sunindextype m, sunindextype n, realtype *beta, realtype *v)
    int denseORMQR(realtype **a, sunindextype m, sunindextype n, realtype *beta,
                   realtype *v, realtype *w, realtype *wrk)
    void DenseCopy(DlsMat A, DlsMat B)
    void denseCopy(realtype **a, realtype **b, sunindextype m, sunindextype n)
    void DenseScale(realtype c, DlsMat A)
    void denseScale(realtype c, realtype **a, sunindextype m, sunindextype n)
    void denseAddIdentity(realtype **a, sunindextype n)
    void DenseMatvec(DlsMat A, realtype *x, realtype *y)
    void denseMatvec(realtype **a, realtype *x, realtype *y, sunindextype m, sunindextype n)

cdef extern from "sundials/sundials_nonlinearsolver.h":

    struct _generic_SUNNonlinearSolver_Ops:
        pass
    struct _generic_SUNNonlinearSolver:
        pass
    ctypedef _generic_SUNNonlinearSolver_Ops *SUNNonlinearSolver_Ops
    ctypedef _generic_SUNNonlinearSolver *SUNNonlinearSolver

    ctypedef int (*SUNNonlinSolSysFn)(N_Vector y, N_Vector F, void* mem)
    ctypedef int (*SUNNonlinSolLSetupFn)(booleantype jbad,
                                         booleantype* jcur, void* mem)
    ctypedef int (*SUNNonlinSolLSolveFn)(N_Vector b, void* mem)
    # rename reserved del into del_t for python!
    ctypedef int (*SUNNonlinSolConvTestFn)(SUNNonlinearSolver NLS, N_Vector y,
                                          N_Vector del_t, realtype tol, 
                                          N_Vector ewt, void* mem)

    cdef enum SUNNonlinearSolver_Type:
        SUNNONLINEARSOLVER_ROOTFIND,
        SUNNONLINEARSOLVER_FIXEDPOINT

    struct _generic_SUNNonlinearSolver_Ops:
        SUNNonlinearSolver_Type (*gettype)(SUNNonlinearSolver)
        int (*initialize)(SUNNonlinearSolver)
        int (*setup)(SUNNonlinearSolver, N_Vector, void*)
        int (*solve)(SUNNonlinearSolver, N_Vector, N_Vector, N_Vector, realtype,
                     booleantype, void*)
        int (*free)(SUNNonlinearSolver)
        int (*setsysfn)(SUNNonlinearSolver, SUNNonlinSolSysFn)
        int (*setlsetupfn)(SUNNonlinearSolver, SUNNonlinSolLSetupFn)
        int (*setlsolvefn)(SUNNonlinearSolver, SUNNonlinSolLSolveFn)
        int (*setctestfn)(SUNNonlinearSolver, SUNNonlinSolConvTestFn)
        int (*setmaxiters)(SUNNonlinearSolver, int)
        int (*getnumiters)(SUNNonlinearSolver, long int*)
        int (*getcuriter)(SUNNonlinearSolver, int*)
        int (*getnumconvfails)(SUNNonlinearSolver, long int*)

    struct _generic_SUNNonlinearSolver :
        void *content
        SUNNonlinearSolver_Ops ops

    SUNNonlinearSolver SUNNonlinSolNewEmpty();
    void SUNNonlinSolFreeEmpty(SUNNonlinearSolver NLS);

    SUNNonlinearSolver_Type SUNNonlinSolGetType(SUNNonlinearSolver NLS)

    int SUNNonlinSolInitialize(SUNNonlinearSolver NLS)
    int SUNNonlinSolSetup(SUNNonlinearSolver NLS, N_Vector y, void* mem)
    int SUNNonlinSolSolve(SUNNonlinearSolver NLS, N_Vector y0, N_Vector y,
                          N_Vector w, realtype tol,
                          booleantype callLSetup, void *mem)
    int SUNNonlinSolFree(SUNNonlinearSolver NLS)

    int SUNNonlinSolSetSysFn(SUNNonlinearSolver NLS,  SUNNonlinSolSysFn SysFn)
    int SUNNonlinSolSetLSetupFn(SUNNonlinearSolver NLS,
                                SUNNonlinSolLSetupFn SetupFn)
    int SUNNonlinSolSetLSolveFn(SUNNonlinearSolver NLS,
                                SUNNonlinSolLSolveFn SolveFn)
    int SUNNonlinSolSetConvTestFn(SUNNonlinearSolver NLS,
                                  SUNNonlinSolConvTestFn CTestFn)
    int SUNNonlinSolSetMaxIters(SUNNonlinearSolver NLS, int maxiters)

    int SUNNonlinSolGetNumIters(SUNNonlinearSolver NLS, long int *niters)
    int SUNNonlinSolGetCurIter(SUNNonlinearSolver NLS, int *iter)
    int SUNNonlinSolGetNumConvFails(SUNNonlinearSolver NLS,
                                    long int *nconvfails)


    enum: SUN_NLS_SUCCESS         #  0
    enum: SUN_NLS_CONTINUE        # +901
    enum: SUN_NLS_CONV_RECVR      # +902
    
    enum: SUN_NLS_MEM_NULL        # -901
    enum: SUN_NLS_MEM_FAIL        # -902
    enum: SUN_NLS_ILL_INPUT       # -903
    enum: SUN_NLS_VECTOROP_ERR    # -904
    enum: SUN_NLS_EXT_FAIL        # -905 



# Authors: B. Malengier, russel (scipy trac)
from __future__ import print_function

"""
Tests for differential algebraic equation solvers.
"""
import numpy
import platform
import scipy.sparse as sparse

from numpy import (arange, zeros, array, dot, sqrt, cos, sin, allclose,
                    empty, alen)

from numpy.testing import TestCase, run_module_suite
from scipy.integrate import ode as Iode
from scikits.odes import ode,dae
from scikits.odes.sundials.common_defs import DTYPE

class TestDae(TestCase):
    """
    Check integrate.dae
    """
    def _do_problem(self, problem, integrator, old_api=True, **integrator_params):
        jac = None
        if hasattr(problem, 'jac'):
            jac = problem.jac
        res = problem.res

        class UserData:
            def __init__(self):
                self.J = None

        my_userdata = UserData()

        jac_times_vec = None
        jac_times_vec2 = None
        jac_times_setupfn = None

        # restrict to 'ida' method since jacobian function below
        # follows the ida interface
        if jac is not None and integrator in ['ida', 'idas']:
            def jac_times_vec(tt, yy, yp, rr, v, Jv, cj):
                J = zeros((len(yy), len(yy)), DTYPE)
                jac(tt, yy, yp, rr, cj, J)
                Js = sparse.csr_matrix(J)
                Jv[:] = Js * v
                return 0

            def jac_times_vec2(tt, yy, yp, rr, v, Jv, cj, userdata):
                Jv[:] = userdata.J * v
                return 0

            def jac_times_setupfn(tt, yy, yp, rr, cj, userdata):
                J = zeros((len(yy), len(yy)), DTYPE)
                jac(tt, yy, yp, rr, cj, J)
                userdata.J = sparse.csr_matrix(J)
                return 0

        igs = [dae(integrator, res, jacfn=jac, old_api=old_api)]

        # if testing 'ida' then try the iterative linsolvers as well
        if integrator in ['ida', 'idas']:
            igs.append(
                dae(integrator, res, linsolver='spgmr',
                    jac_times_vecfn=jac_times_vec,
                    old_api=old_api)
            )
            igs.append(
                dae(integrator, res, linsolver='spgmr',
                    jac_times_vecfn=jac_times_vec2,
                    jac_times_setupfn=jac_times_setupfn,
                    old_api=old_api,
                    user_data=my_userdata)
            )

        for ig in igs:
            ig.set_options(old_api=old_api, **integrator_params)
            z = empty((1+len(problem.stop_t),alen(problem.z0)), DTYPE)
            zprime = empty((1+len(problem.stop_t),alen(problem.z0)), DTYPE)
            ist = ig.init_step(0., problem.z0, problem.zprime0, z[0], zprime[0])
            i=1
            for time in problem.stop_t:
                soln = ig.step(time, z[i], zprime[i])
                if old_api:
                    flag, rt = soln
                else:
                    flag = soln.flag
                    rt = soln.values.t
                i += 1
                if integrator in ['ida', 'idas']:
                    assert flag==0, (problem.info(), flag)
                else:
                    assert flag > 0, (problem.info(), flag)

            assert problem.verify(array(z), array(zprime),  [0.]+problem.stop_t), \
                        (problem.info(),)


    def test_ddaspk(self):
        """Check the ddaspk solver"""
        for problem_cls in PROBLEMS_DDASPK:
            problem = problem_cls()
            self._do_problem(problem, 'ddaspk', **problem.ddaspk_pars)

    def test_lsodi(self):
        """Check the lsodi solver"""
        # skip on aarch64, see https://github.com/bmcage/odes/issues/101
        if platform.machine() == "aarch64":
            return
        for problem_cls in PROBLEMS_LSODI:
            problem = problem_cls()
            self._do_problem(problem, 'lsodi', **problem.lsodi_pars)

    def test_ida_old_api(self):
        """Check the ida solver"""
        for problem_cls in PROBLEMS_IDA:
            problem = problem_cls()
            self._do_problem(problem, 'ida', old_api=True, **problem.ida_pars)

    def test_ida(self):
        """Check the ida solver"""
        for problem_cls in PROBLEMS_IDA:
            problem = problem_cls()
            self._do_problem(problem, 'ida', old_api=False, **problem.ida_pars)

    def test_idas(self):
        """Check that idas solver behaves as ida"""
        for problem_cls in PROBLEMS_IDA:
            problem = problem_cls()
            self._do_problem(problem, 'idas', **problem.ida_pars)

#------------------------------------------------------------------------------
# Test problems
#------------------------------------------------------------------------------

def simple_adda(t,y,ml,mu,p,nrowp):
    p[0,0] += 1.0
    p[1,1] += 1.0
    return p

class DAE:
    """
    DAE problem
    """
    stop_t  = [1]
    z0      = []
    zprime0 =  []

    atol    = 1e-6
    rtol    = 1e-5

    ddaspk_pars = {}
    ida_pars = {}
    lsodi_pars = {'adda' : simple_adda}

    def info(self):
        return self.__class__.__name__ + ": No info given"

class SimpleOscillator(DAE):
    r"""
    Free vibration of a simple oscillator::
        m \ddot{u} + k u = 0, u(0) = u_0, \dot{u}(0)=\dot{u}_0
    Solution::
        u(t) = u_0*cos(sqrt(k/m)*t)+\dot{u}_0*sin(sqrt(k/m)*t)/sqrt(k/m)
    """
    stop_t  = [2 + 0.09, 3]
    u0      = 1.
    dotu0   = 0.1

    k = 4.0
    m = 1.0
    z0      = array([dotu0, u0], DTYPE)
    zprime0 = array([-k*u0/m, dotu0], DTYPE)

    def __init__(self):
        self.lsodi_pars = {'adda_func' : self.adda}

    def info(self):
        doc = self.__class__.__name__ + ": 2x2 constant mass matrix"
        return doc

    def res(self, t, z, zp, res):
        tmp1 = zeros((2,2), DTYPE)
        tmp2 = zeros((2,2), DTYPE)
        tmp1[0,0] = self.m
        tmp1[1,1] = 1.
        tmp2[0,1] = self.k
        tmp2[1,0] = -1.
        res[:] = dot(tmp1, zp)[:]+dot(tmp2, z)[:]

    def adda(self, t, y, ml, mu, p, nrowp):
        p[0,0] -= self.m
        p[1,1] -= 1.0
        return p

    def verify(self, zs, zps, t):
        omega = sqrt(self.k / self.m)
        ok = True
        for (z, zp, time) in zip(zs, zps, t):
            u = self.z0[1]*cos(omega*time)+self.z0[0]*sin(omega*time)/omega
            ok = ok and allclose(u, z[1], atol=self.atol, rtol=self.rtol) and \
            allclose(z[0], zp[1], atol=self.atol, rtol=self.rtol)
            ##print('verify SO', time, ok, z, u, self.z0)
        return ok

class SimpleOscillatorJac(SimpleOscillator):
    def jac(self, t, y, yp, cj, jac):
        """Jacobian[i,j] is dRES(i)/dY(j) + CJ*dRES(i)/dYPRIME(j)"""
        jc = zeros((len(y), len(y)), DTYPE)
        cj_in = cj
        jac[0][0] = self.m*cj_in ;jac[0][1] = self.k
        jac[1][0] = -1       ;jac[1][1] = cj_in;

class SimpleOscillatorJacIDA(SimpleOscillator):
    def jac(self, t, y, yp, residual, cj, jac):
        """Jacobian[i,j] is dRES(i)/dY(j) + CJ*dRES(i)/dYPRIME(j)"""
        jc = zeros((len(y), len(y)), DTYPE)
        cj_in = cj
        jac[0][0] = self.m*cj_in ;jac[0][1] = self.k
        jac[1][0] = -1       ;jac[1][1] = cj_in;

class StiffVODECompare(DAE):
    r"""
    We create a stiff problem, obtain the vode solution, and compare with
    dae solution
    Correct solution with runga-kutta 45:
     [t = 0.4, y0(t) = 0.985172121250895372, y1(t) = 0.0000338791735424692934,
               y2(t) = 0.0147939995755618956]
     [t = 4., y0(t) = 0.905519130228504610, y1(t) = 0.0000224343714361267687,
               y2(t) = 0.0944584354000592570]
    """
    z0      = array([1., 0., 0.], DTYPE)
    zprime0 = array([-0.04, 0.04, 0.], DTYPE)

    atol    = 1e-4
    rtol    = 1e-4

    def f_vode(self, t, y):
        ydot0 = -0.04*y[0] + 1e4*y[1]*y[2]
        ydot2 = 3e7*y[1]*y[1]
        ydot1 = -ydot0-ydot2
        return array([ydot0,ydot1,ydot2])

    def jac_vode(self, t, y):
        jc = [[-0.04,1e4*y[2]          ,1e4*y[1]],
              [0.04 ,-1e4*y[2]-6e7*y[1],-1e4*y[1]],
              [0.0    ,6e7*y[1]           ,0.0]]
        return array(jc)

    def f_cvode(self, t, y, ydot):
        ydot[:] = self.f_vode(t, y)

    def jac_cvode(self, t, y, fy, J):
        J[:,:] = self.jac_vode(t, y)

    def __init__(self):
        """We obtain the vode solution first"""
        r = Iode(self.f_vode, self.jac_vode).set_integrator('vode',
                                  rtol=[1e-4,1e-4,1e-4],
                                  atol=[1e-8,1e-14,1e-6],
                                  method='bdf',
                                  )
        r.set_initial_value([1.,0.,0.])
        nr = 5
        self.sol = zeros((nr, 3))
        self.stop_t = [0., 0.4, 4., 40., 400.]
        self.sol[0] = r.y
        i=1
        for time in self.stop_t[1:]:
            r.integrate(time)
            self.sol[i] = r.y
            i +=1
        #we now do the same with the sundials CVODE solution
        r = ode('cvode', self.f_cvode, jacfn=self.jac_cvode,
                         rtol=1e-4,
                         atol=[1e-8,1e-14,1e-6],
                         lmm_type='bdf',
                         old_api=False,
        )
        soln = r.solve(self.stop_t, [1.,0.,0.])
        self.sol2 = soln.values.y

        #for the solvers, only stop time, not start:
        self.stop_t = self.stop_t[1:]

        #we need to activate some extra parameters in the solver
        #order par is rtol,atol,lband,uband,tstop,order,nsteps,
        #         max_step,first_step,enforce_nonnegativity,nonneg_type,
        #         compute_initcond,constraint_init,constraint_type,algebraic_var
        self.ddaspk_pars = {'rtol' : [1e-4,1e-4,1e-4],
                            'atol' : [1e-8,1e-14,1e-6],
                           }
        self.ida_pars = {'rtol' : 1e-4,
                         'atol' : [1e-8,1e-14,1e-6],
                        }
        self.lsodi_pars = {'rtol' : [1e-4,1e-4,1e-4],
                            'atol' : [1e-6,1e-10,1e-6],
                            'adda_func' : self.adda
                           }

    def res(self, t, y, yp, res):
        res[0] = yp[0] + 0.04*y[0] - 1e4*y[1]*y[2]
        res[2] = yp[2] - 3e7*y[1]*y[1]
        res[1] = yp[1] +yp[0]+yp[2]

    def adda(self, t, y, ml, mu, p, nrowp):
        p[0,0] -= 1.0
        p[1,0] -= 1.0
        p[1,1] -= 1.0
        p[1,2] -= 1.0
        p[2,2] -= 1.0
        return p

    def verify(self, y, yp, t):
        return ( allclose(self.sol, y, atol=self.atol, rtol=self.rtol) and
                 allclose(self.sol2, y, atol=self.atol, rtol=self.rtol) )

PROBLEMS_DDASPK = [SimpleOscillator, StiffVODECompare,
            SimpleOscillatorJac ]
PROBLEMS_IDA = [SimpleOscillator, StiffVODECompare,
            SimpleOscillatorJacIDA ]
PROBLEMS_IDAS = [SimpleOscillator, StiffVODECompare,
            SimpleOscillatorJacIDA ]
PROBLEMS_LSODI = [SimpleOscillator, StiffVODECompare]
#------------------------------------------------------------------------------

if __name__ == "__main__":
    try:
        run_module_suite()
    except NameError:
        test = TestDae()
        test.test_ddaspk()
        test.test_lsodi()
        test.test_ida_old_api()
        test.test_ida()



# -*- coding: utf-8 -*-
# Authors: B. Malengier based on ode.py
r"""
First-order DAE solver
======================

User-friendly interface to various numerical integrators for solving an
algebraic system of first order ODEs with prescribed initial conditions:

.. math::
    A \frac{dy(t)}{dt} = f(t,y(t)),

    y(t=0)[i] = y0[i],

    \frac{d y(t=0)}{dt}[i]  = yprime0[i],

where :math:`i = 0, ..., len(y0) - 1`; :math:`A` is a (possibly singular) matrix
of size :math:`i × i`; and :math:`f(t,y)` is a vector of size :math:`i` or more generally, equations of the form

.. math::
    G(t,y,y') = 0

"""

from __future__ import print_function

__all__ = ['dae']
__version__ = "$Id$"
__docformat__ = "restructuredtext en"

from numpy import asarray, array, zeros, sin, int32, isscalar, empty, alen
from copy import copy
import re, sys

class DaeBase(object):
    """
    The interface which DAE solvers must implement.

    Parameters
    ----------
    Rfn :
        residual function
    options : mapping
        Additional options for initialization, solver dependent
    """

    integrator_classes = []

    def __init__(self, Rfn, **options):
        raise NotImplementedError('all DAE solvers must implement this')

    def set_options(self, **options):
        """
        Set specific options for the solver.

        Calling set_options a second time, normally resets the solver.
        """
        raise NotImplementedError('all DAE solvers must implement this')

    def solve(self, tspan, y0,  yp0):
        """
        Runs the solver.

        Parameters
        ----------
            tspan : list/array
                A list of times at which the computed value will be returned. Must contain the start time as first entry.
            y0 : list/array
                List of initial values
            yp0 : list/array
                List of initial values of derivatives

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnumXXX)
            ``values``  Named tuple with entries array t and array y and array ydot. y will correspond to y_retn value and ydot to yp_retn!
            ``errors``  Named tuple with entries t and y and ydot of error
            ``roots``   Named tuple with entries array t and array y and array ydot
            ``tstop``   Named tuple with entries array t and array y and array ydot
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   indicating return status of the solver
            ``t``      numpy array of times at which the computations were successful
            ``y``      numpy array of values corresponding to times t (values of y[i, :] ~ t[i])
            ``yp``     numpy array of derivatives corresponding to times t (values of yp[i, :] ~ t[i])
            ``t_err``  float or None - if recoverable error occured (for example reached maximum number of allowed iterations), this is the time at which it happened
            ``y_err``  numpy array of values corresponding to time t_err
            ``yp_err`` numpy array of derivatives corresponding to time t_err
            ========== ==========================================

        """
        raise NotImplementedError('all DAE solvers must implement this')

    def init_step(self, t0, y0, yp0, y_ic0_retn = None, yp_ic0_retn = None):
        """
        Initializes the solver and allocates memory.

        Parameters
        ----------
        t0 : number
            initial time
        y0 : list/array
            initial condition for y
        yp0 :  list/array
            initial condition for yp
        y_ic0 : numpy array
            (optional) returns the calculated consistent initial condition for y
        yp_ic0 :  numpy array
            (optional) returns the calculated consistent initial condition for y derivated.

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnumXXX)
            ``values``  Named tuple with entries t and y and ydot. y will correspond to y_retn value and ydot to yp_retn!
            ``errors``  Named tuple with entries t and y and ydot
            ``roots``   Named tuple with entries t and y and ydot
            ``tstop``   Named tuple with entries t and y and ydot
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    status of the computation (successful or error occured)
            ``t_out``   time, where the solver stopped (when no error occured, t_out == t)
            =========== ==========================================

        """
        raise NotImplementedError('all DAE solvers must implement this')

    def step(self, t, y_retn=None, yp_retn=None):
        """
        Method for calling successive next step of the IDA solver to allow
        more precise control over the IDA solver. The 'init_step' method has to
        be called before the 'step' method.

        A step is done towards time t, and output at t returned. This time can be higher or lower than the previous time. If option 'one_step_compute'==True, and the solver supports it, only one internal solver step is done in the direction of t starting at the current step.

        If old_api=True, the old behavior is used: if t>0.0 then integration is
        performed until this time and results at this time are returned in
        y_retn; else if if t<0.0 only one internal step is perfomed towards time
        abs(t) and results after this one time step are returned.

        Parameters
        ----------
        t : number
        y_retn : numpy array (ndim = 1) or None.
            (Needs to be preallocated) If not None, will be filled with y at time t. If None y_retn is not used.
        yp_retn : numpy array (ndim = 1) or None.
            (Needs to be preallocated) If not None, will be filled with derivatives of y at time t. If None yp_retn is not used.

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnumXXX)
            ``values``  Named tuple with entries t and y and ydot. y will correspond to y_retn value and ydot to yp_retn!
            ``errors``  Named tuple with entries t and y and ydot
            ``roots``   Named tuple with entries t and y and ydot
            ``tstop``   Named tuple with entries t and y and ydot
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    status of the computation (successful or error occured)
            ``t_out``   time, where the solver stopped (when no error occured, t_out == t)
            =========== ==========================================

        """
        raise NotImplementedError('all DAE solvers must implement this')

#------------------------------------------------------------------------------
# User interface
#------------------------------------------------------------------------------

class dae(object):
    """
    A generic interface class to differential algebraic equations.

    Define equation res = G(t,y,y') which can eg be G = f(y,t) - A y' when solving A y' = f(y,t), and where (optional) jac is the jacobian matrix of the nonlinear system see fortran source code), so d res/dy + scaling * d res/dy' or d res/dy depending on the backend.

    Parameters
    ----------
    integrator_name : ``'ida'``, ``'ddaspk'`` or ``'lsodi'``
        The integrator solver to use.

    eqsres : residual function
        Residual of the DAE. The signature of this function depends on the
        solver used, see the solver documentation for details.
        Generally however, you can assume the following signature to work:

            ``eqsres(x, y, yprime, return_residual)``

        with
        x       : independent variable, eg the time, float
        y       : array of n unknowns in x
        yprime  : dy/dx array of n unknowns in x, dimension = dim(y)
        return_residual: array that must be updated with the value of the residuals, so G(t,y,y').  The dimension is equal to dim(y)
        return value: integer, 0 for success. It is not guaranteed that a solver takes this status into account

        Some solvers will allow userdata to be passed to eqsres, or optional
        formats that are more performant.

    options : mapping
        Additional options for initialization, solver dependent
        See set_options method of the `integrator_name` you selected for
        details.


    See Also
    --------
    odeint : an ODE integrator with a simpler interface based on lsoda from ODEPACK
    ode : class around vode ODE integrator


    Notes
    -----
    Possible future solvers

    ddaskr: Not included, starting hints: http://osdir.com/ml/python.f2py.user/2005-07/msg00014.html
    Modified Extended Backward Differentiation Formulae (MEBDF): Not included. Fortran codes: http://www.ma.ic.ac.uk/~jcash/IVP_software/readme.html

    Examples
    --------
    DAE arise in many applications of dynamical systems, as well as in
    discritisations of PDE (eg moving mesh combined with method of
    lines).
    As an easy example, consider the simple oscillator, which we write as
    G(y,y',t) = 0 instead of the normal ode, and solve as a DAE.

    >>> from __future__ import print_function
    >>> from numpy import cos, sin, sqrt
    >>> k = 4.0
    >>> m = 1.0
    >>> initx = [1, 0.1]
    >>> initxp = [initx[1], -k/m*initx[0]]
    >>> def reseqn(t, x, xdot, result):
        ... # we create residual equations for the problem
        ... result[0] = m*xdot[1] + k*x[0]
        ... result[1] = xdot[0] - x[1]
    >>> from scikits.odes import dae
    >>> solver = dae('ida', reseqn)
    >>> result = solver.solve([0., 1., 2.], initx, initxp)
    """

    LOADED = False

    def __init__(self, integrator_name, eqsres, **options):
        integrator = find_dae_integrator(integrator_name)
        if integrator is None:
            raise ValueError('No integrator name match with %s or is not available.'\
                  %(repr(integrator_name)))
        else:
            self._integrator = integrator(eqsres, **options)

    def set_options(self, **options):
        """
        Set specific options for the solver.
        See the solver documentation for details.

        Calling set_options a second time, normally resets the solver.
        """
        return self._integrator.set_options(**options)

    def solve(self, tspan, y0,  yp0):
        """
        Runs the solver.

        Parameters
        ----------
        tspan : list/array
            A list of times at which the computed value will be returned. Must contain the start time as first entry.
        y0 : list/array
            list array of initial values
        yp0 : list/array
            list array of initial values of derivatives

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnumXXX)
            ``values``  Named tuple with entries array t and array y and array ydot. y will correspond to y_retn value and ydot to yp_retn!
            ``errors``  Named tuple with entries t and y and ydot of error
            ``roots``   Named tuple with entries array t and array y and array ydot
            ``tstop``   Named tuple with entries array t and array y and array ydot
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   indicating return status of the solver
            ``t``      numpy array of times at which the computations were successful
            ``y``      numpy array of values corresponding to times t (values of y[i, :] ~ t[i])
            ``yp``     numpy array of derivatives corresponding to times t (values of yp[i, :] ~ t[i])
            ``t_err``  float or None - if recoverable error occured (for example reached maximum number of allowed iterations), this is the time at which it happened
            ``y_err``  numpy array of values corresponding to time t_err
            ``yp_err`` numpy array of derivatives corresponding to time t_err
            ========== ==========================================

        """
        return self._integrator.solve(tspan, y0,  yp0)

    def init_step(self, t0, y0, yp0, y_ic0_retn = None, yp_ic0_retn = None):
        """
        Initializes the solver and allocates memory. It is not needed to
        call this method if solve is used to compute the solution. In the case
        step is used, init_step must be called first.

        Parameters
        ----------
        t0 : number
            initial time
        y0 : list/array
            initial condition for y
        yp0 :  list/array
            initial condition for yp
        y_ic0 : numpy array
            (optional) returns the calculated consistent initial condition for y
        yp_ic0 :  numpy array
            (optional) returns the calculated consistent initial condition for y derivated.

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnumXXX)
            ``values``  Named tuple with entries t and y and ydot. y will correspond to y_retn value and ydot to yp_retn!
            ``errors``  Named tuple with entries t and y and ydot
            ``roots``   Named tuple with entries t and y and ydot
            ``tstop``   Named tuple with entries t and y and ydot
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    status of the computation (successful or error occured)
            ``t_out``   time, where the solver stopped (when no error occured, t_out == t)
            =========== ==========================================

        """
        return self._integrator.init_step(t0, y0, yp0, y_ic0_retn, yp_ic0_retn)

    def step(self, t, y_retn=None, yp_retn=None):
        """
        Method for calling successive next step of the IDA solver to allow
        more precise control over the IDA solver. The 'init_step' method has to
        be called before the 'step' method.

        A step is done towards time t, and output at t returned. This time can be higher or lower than the previous time. If option 'one_step_compute'==True, and the solver supports it, only one internal solver step is done in the direction of t starting at the current step.

        If old_api=True, the old behavior is used: if t>0.0 then integration is
        performed until this time and results at this time are returned in
        y_retn; else if if t<0.0 only one internal step is perfomed towards time
        abs(t) and results after this one time step are returned.

        Parameters
        ----------
        t : number
        y_retn : numpy array (ndim = 1) or None.
            (Needs to be preallocated) If not None, will be filled with y at time t. If None y_retn is not used.
        yp_retn : numpy array (ndim = 1) or None.
            (Needs to be preallocated) If not None, will be filled with derivatives of y at time t. If None yp_retn is not used.

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnumXXX)
            ``values``  Named tuple with entries t and y and ydot. y will correspond to y_retn value and ydot to yp_retn!
            ``errors``  Named tuple with entries t and y and ydot
            ``roots``   Named tuple with entries t and y and ydot
            ``tstop``   Named tuple with entries t and y and ydot
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    status of the computation (successful or error occured)
            ``t_out``   time, where the solver stopped (when no error occured, t_out == t)
            =========== ==========================================

        """
        return self._integrator.step(t, y_retn, yp_retn)

    def __del__(self):
        """
        Clean up what is needed
        """
        if hasattr(self, '_integrator'):
            del self._integrator

#------------------------------------------------------------------------------
# DAE integrators
#------------------------------------------------------------------------------

def find_dae_integrator(name):
    if not dae.LOADED:
        ## ida
        try:
            from .sundials import ida
            DaeBase.integrator_classes.append(ida.IDA)
        except ValueError as msg:
            print('Could not load IDA solver', msg)
        except ImportError:
            print(sys.exc_info()[1])

        ## idas
        try:
            from .sundials import idas
            DaeBase.integrator_classes.append(idas.IDAS)
        except ValueError as msg:
            print('Could not load IDAS solver', msg)
        except ImportError:
            print(sys.exc_info()[1])

        ## ddaspk
        try:
            from .ddaspkint import ddaspk
        except ImportError:
            print(sys.exc_info()[1])

        ## lsodi
        try:
            from .lsodiint import lsodi
        except ImportError:
            print(sys.exc_info()[1])

        dae.LOADED = True

    for cl in DaeBase.integrator_classes:
        if re.match(name, cl.__name__, re.I):
            return cl
        elif hasattr(cl, name) and re.match(name, cl.name, re.I):
            return cl
    raise ValueError('Integrator name %s does not exsist' % name)



# -*- coding: utf-8 -*-
"""
Created on Tue Feb  2 08:54:10 2016

@author: benny

based on the test in scipy.integrate:
# Authors: Nils Wagner, Ed Schofield, Pauli Virtanen, John Travers

Tests for odeint
"""
from __future__ import division, print_function, absolute_import

import numpy as np
from numpy import (arange, zeros, array, dot, sqrt, cos, sin, eye, pi, exp,
                   allclose)

from numpy.testing import (
    assert_, TestCase, run_module_suite, assert_array_almost_equal,
    assert_raises, assert_allclose, assert_array_equal, assert_equal)

from scikits.odes.odeint import odeint
from scikits.odes.sundials.common_defs import DTYPE, has_lapack

# Lapack only compatible with precision double and index type 32.
TEST_LAPACK = (DTYPE == np.double and has_lapack)

#------------------------------------------------------------------------------
# Test ODE integrators
#------------------------------------------------------------------------------

class TestOdeint(TestCase):
    # Check integrate.odeint
    def _do_problem(self, problem):
        t = arange(0.0, problem.stop_t, 0.05)
        for method in ['bdf', 'admo', 'rk5', 'rk8']:
            sol = odeint(problem.f, t, problem.z0, method=method)
            assert_(problem.verify(sol.values.y, sol.values.t),
                    msg="Method {} on Problem {}. Printout {}".format(method, 
                            problem.__class__.__name__,
                            False
                            #problem.verify(sol.values.y, sol.values.t, True)
                            )
                    )

    def test_odeint(self):
        for problem_cls in PROBLEMS:
            problem = problem_cls()
            if problem.cmplx:
                continue
            self._do_problem(problem)

#------------------------------------------------------------------------------
# Test problems
#------------------------------------------------------------------------------

class ODE:
    """
    ODE problem
    """
    stiff = False
    cmplx = False
    stop_t = 1
    z0 = []

    lband = None
    uband = None

    atol = 1e-4 #1e-6
    rtol = 1e-6 #1e-5


class SimpleOscillator(ODE):
    r"""
    Free vibration of a simple oscillator::
        m \ddot{u} + k u = 0, u(0) = u_0 \dot{u}(0) \dot{u}_0
    Solution::
        u(t) = u_0*cos(sqrt(k/m)*t)+\dot{u}_0*sin(sqrt(k/m)*t)/sqrt(k/m)
    """
    stop_t = 1 + 0.09
    z0 = array([1.0, 0.1], float)

    k = 4.0
    m = 1.0

    def f(self, t, z, rhs):
        tmp = np.zeros((2,2), float)
        tmp[0, 1] = 1.0
        tmp[1, 0] = -self.k / self.m
        rhs[:] = dot(tmp, z)
        return 0

    def verify(self, zs, t, printout=False):
        omega = sqrt(self.k / self.m)
        u = self.z0[0]*cos(omega*t) + self.z0[1]*sin(omega*t)/omega
        if printout:
            print ('exact=', u, ', computed=', zs[:, 0])
        return allclose(u, zs[:, 0], atol=self.atol, rtol=self.rtol)

class ComplexExp(ODE):
    r"""The equation :lm:`\dot u = i u`"""
    stop_t = 1.23*pi
    z0 = exp([1j, 2j, 3j, 4j, 5j])
    cmplx = True

    def f(self, t, z, rhs):
        rhs[:] = 1j*z
        return 0

    def jac(self, t, z, fz, J):
        J[:,:] = 1j*eye(5)
        return 0

    def verify(self, zs, t):
        u = self.z0 * exp(1j*t)
        return allclose(u, zs, atol=self.atol, rtol=self.rtol)


class Pi(ODE):
    r"""Integrate 1/(t + 1j) from t=-10 to t=10"""
    stop_t = 20
    z0 = [0]
    cmplx = True

    def f(self, t, z, rhs):
        rhs[:] = array([1./(t - 10 + 1j)])
        return 0

    def verify(self, zs, t, printout=False):
        u = -2j * np.arctan(10)
        return allclose(u, zs[-1, :], atol=self.atol, rtol=self.rtol)


class CoupledDecay(ODE):
    r"""
    3 coupled decays suited for banded treatment
    (banded mode makes it necessary when N>>3)
    """

    stiff = True
    stop_t = 0.5
    z0 = [5.0, 7.0, 13.0]
    lband = 1
    uband = 0

    lmbd = [0.17, 0.23, 0.29]  # fictious decay constants

    def f(self, t, z, rhs):
        lmbd = self.lmbd
        rhs[:] = np.array([-lmbd[0]*z[0],
                         -lmbd[1]*z[1] + lmbd[0]*z[0],
                         -lmbd[2]*z[2] + lmbd[1]*z[1]])
        return 0

    def jac(self, t, z, fz, J):
        # The full Jacobian is
        #
        #    [-lmbd[0]      0         0   ]
        #    [ lmbd[0]  -lmbd[1]      0   ]
        #    [    0      lmbd[1]  -lmbd[2]]
        #
        # The lower and upper bandwidths are lband=1 and uband=0, resp.
        # The representation of this array in packed format is
        #
        #    [-lmbd[0]  -lmbd[1]  -lmbd[2]]
        #    [ lmbd[0]   lmbd[1]      0   ]

        lmbd = self.lmbd
        j = np.zeros((self.lband + self.uband + 1, 3), order='F')

        def set_j(ri, ci, val):
            j[self.uband + ri - ci, ci] = val
        set_j(0, 0, -lmbd[0])
        set_j(1, 0, lmbd[0])
        set_j(1, 1, -lmbd[1])
        set_j(2, 1, lmbd[1])
        set_j(2, 2, -lmbd[2])
        J[:,:] = j
        return 0

    def verify(self, zs, t, printout=False):
        # Formulae derived by hand
        lmbd = np.array(self.lmbd)
        d10 = lmbd[1] - lmbd[0]
        d21 = lmbd[2] - lmbd[1]
        d20 = lmbd[2] - lmbd[0]
        e0 = np.exp(-lmbd[0] * t)
        e1 = np.exp(-lmbd[1] * t)
        e2 = np.exp(-lmbd[2] * t)
        u = np.vstack((
            self.z0[0] * e0,
            self.z0[1] * e1 + self.z0[0] * lmbd[0] / d10 * (e0 - e1),
            self.z0[2] * e2 + self.z0[1] * lmbd[1] / d21 * (e1 - e2) +
            lmbd[1] * lmbd[0] * self.z0[0] / d10 *
            (1 / d20 * (e0 - e2) - 1 / d21 * (e1 - e2)))).transpose()
        return allclose(u, zs, atol=self.atol, rtol=self.rtol)


PROBLEMS = [SimpleOscillator, ComplexExp, Pi, CoupledDecay]

#------------------------------------------------------------------------------


def f(t, x):
    dxdt = [x[1], -x[0]]
    return dxdt


def jac(t, x):
    j = array([[0.0, 1.0],
               [-1.0, 0.0]])
    return j


def f1(t, x, omega):
    dxdt = [omega*x[1], -omega*x[0]]
    return dxdt


def jac1(t, x, omega):
    j = array([[0.0, omega],
               [-omega, 0.0]])
    return j


def f2(t, x, omega1, omega2):
    dxdt = [omega1*x[1], -omega2*x[0]]
    return dxdt


def jac2(t, x, omega1, omega2):
    j = array([[0.0, omega1],
               [-omega2, 0.0]])
    return j


def fv(t, x, omega):
    dxdt = [omega[0]*x[1], -omega[1]*x[0]]
    return dxdt


def jacv(t, x, omega):
    j = array([[0.0, omega[0]],
               [-omega[1], 0.0]])
    return j

def test_odeint_trivial_time():
    # Test that odeint gives ValueError if only a single time point is given
    y0 = 1
    t = [0]
    for method in ['bdf', 'admo', 'rk5', 'rk8']:
        #sol = odeint(lambda t, y: -y, t, y0, method=method)
        assert_raises(ValueError, odeint, lambda t, y: -y, t, y0)

c = array([[-205, 0.01, 0.00, 0.0],
           [0.1, -2.50, 0.02, 0.0],
           [0.00, 0.01, -2.0, 0.01],
           [0.00, 0.00, 0.1, -1.0]])

cband1 = array([
           [0.  ,  0.01, 0.02,  0.01],
           [-205, -2.50, -2.0, -1.0 ],
           [0.1,   0.01,  0.1,  0.0 ],
           [0.  ,    0., 0.00,  0.0 ]])
cband2 = array([[0, 0.01, 0.02, 0.01],
           [-250, -2.50, -2., -1.],
           [0.1, 0.01, 0.1, 0.]])

def test_odeint_banded_jacobian():
    # Test the use of the `Dfun`, `ml` and `mu` options of odeint default bdf
    # solver

    def func(t, y, rhs):
        rhs[:] = c.dot(y)
        return 0

    def jac(t, y, fy, J):
        J[:,:] = c
        return 0

    def bjac1_rows(t, y, fy, J):
        J[:,:] = cband1
        return 0

    def bjac2_rows(t, y, fy, J):
        # define BAND_ELEM(A,i,j) ((A->cols)[j][(i)-(j)+(A->s_mu)])
        J[:,:] = cband2
        return 0

    y0 = np.ones(4)
    t = np.array([0, 5, 10, 100])

    # Use the full Jacobian.
    sol1 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         jacfn=jac)

    # Use the full Jacobian.
    sol2 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         linsolver='dense')
    # Use the banded Jacobian.
    sol3 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         jacfn=bjac1_rows, lband=2, uband=1, linsolver='band')

    # Use the banded Jacobian.
    sol4 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         jacfn=bjac2_rows, lband=1, uband=1, linsolver='band')

    # Use the banded Jacobian.
    sol5 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         lband=2, uband=1, linsolver='band')

    # Use the banded Jacobian.
    sol6 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         lband=1, uband=1, linsolver='band')
    # Use the diag Jacobian.
    sol7 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         linsolver='diag')

    #use lapack versions:
    if TEST_LAPACK:
        # Use the full Jacobian.
        sol2a = odeint(func, t, y0,
                             atol=1e-13, rtol=1e-11, max_steps=10000,
                             linsolver='lapackdense')
        # Use the banded Jacobian.
        sol3a = odeint(func, t, y0,
                             atol=1e-13, rtol=1e-11, max_steps=10000,
                             jacfn=bjac1_rows, lband=2, uband=1, linsolver='lapackband')

        # Use the banded Jacobian.
        sol4a = odeint(func, t, y0,
                             atol=1e-13, rtol=1e-11, max_steps=10000,
                             jacfn=bjac2_rows, lband=1, uband=1, linsolver='lapackband')

        # Use the banded Jacobian.
        sol5a = odeint(func, t, y0,
                             atol=1e-13, rtol=1e-11, max_steps=10000,
                             lband=2, uband=1, linsolver='lapackband')

        # Use the banded Jacobian.
        sol6a = odeint(func, t, y0,
                             atol=1e-13, rtol=1e-11, max_steps=10000,
                             lband=1, uband=1, linsolver='lapackband')

    #finish with some other solvers
    sol10 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         linsolver='spgmr')
    sol11 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         linsolver='spbcgs')
    sol12 = odeint(func, t, y0,
                         atol=1e-13, rtol=1e-11, max_steps=10000,
                         linsolver='sptfqmr')

    assert_allclose(sol1.values.y, sol2.values.y, atol=1e-12, err_msg="sol1 != sol2")
    assert_allclose(sol1.values.y, sol3.values.y, atol=1e-12, err_msg="sol1 != sol3")
    assert_allclose(sol1.values.y, sol4.values.y, atol=1e-12, err_msg="sol1 != sol4")
    assert_allclose(sol1.values.y, sol5.values.y, atol=1e-12, err_msg="sol1 != sol5")
    assert_allclose(sol1.values.y, sol6.values.y, atol=1e-12, err_msg="sol1 != sol6")
    assert_allclose(sol1.values.y, sol7.values.y, atol=1e-12, err_msg="sol1 != sol7")
    assert_allclose(sol1.values.y, sol10.values.y, atol=1e-12, err_msg="sol1 != sol10")
    assert_allclose(sol1.values.y, sol11.values.y, atol=1e-12, err_msg="sol1 != sol11")
    assert_allclose(sol1.values.y, sol12.values.y, atol=1e-12, err_msg="sol1 != sol12")

    if TEST_LAPACK:
        assert_allclose(sol1.values.y, sol2a.values.y, atol=1e-12, err_msg="sol1 != sol2a")
        assert_allclose(sol1.values.y, sol3a.values.y, atol=1e-12, err_msg="sol1 != sol3a")
        assert_allclose(sol1.values.y, sol4a.values.y, atol=1e-12, err_msg="sol1 != sol4a")
        assert_allclose(sol1.values.y, sol5a.values.y, atol=1e-12, err_msg="sol1 != sol5a")
        assert_allclose(sol1.values.y, sol6a.values.y, atol=1e-12, err_msg="sol1 != sol6a")


def test_odeint_errors():
    def sys1d(t, x, rhs):
        rhs[:] = -100*x
        return 0

    def bad1(t, x, rhs):
        rhs[:] = 1.0/0
        return 0

    def bad2(t, x, rhs):
        rhs[:] = "foo"
        return 0

    def bad_jac1(t, x, fx, J):
        J[:,:] = 1.0/0
        return 0

    def bad_jac2(t, x, fx, J):
        J[:,:] = [["foo"]]
        return 0

    def sys2d(t, x, rhs):
        rhs[:] = [-100*x[0], -0.1*x[1]]
        return 0

    def sys2d_bad_jac(t, x, fx, J):
        J[:,:] = [[1.0/0, 0], [0, -0.1]]
        return 0

    assert_raises(ZeroDivisionError, odeint, bad1, [0, 1], [1.0])

    assert_raises(ValueError, odeint, bad2, [0, 1], [1.0])

    assert_raises(ZeroDivisionError, odeint, sys1d, [0, 1], [1.0], jacfn=bad_jac1)
    assert_raises(ValueError, odeint, sys1d, [0, 1], [1.0], jacfn=bad_jac2)

    assert_raises(ZeroDivisionError, odeint, sys2d, [0, 1], [1.0, 1.0],
                  jacfn=sys2d_bad_jac)

def test_odeint_bad_shapes():
    # Tests of some errors that can occur with odeint.

    def badrhs(t, x, rhs):
        rhs[:] = [1, -1]
        return 0

    def sys1(t, x, rhs):
        rhs[:] = -100*x
        return 0

    def badjac(t, x, fx, J):
        J[:,:] = [[0, 0, 0]]
        return 0

    # y0 must be at most 1-d.
    bad_y0 = [[0, 0], [0, 0]]
    assert_raises(ValueError, odeint, sys1, [0, 1], bad_y0)

    # t must be at most 1-d.
    bad_t = [[0, 1], [2, 3]]
    assert_raises(ValueError, odeint, sys1, bad_t, [10.0])

    # y0 is 10, should be an array [10]
    assert_raises(ValueError, odeint, badrhs, [0, 1], 10)
    # y0 is 10, but badrhs(x, t) returns [1, -1].
    assert_raises(ValueError, odeint, badrhs, [0, 1], [10])

    # shape of array returned by badjac(x, t) is not correct.
    assert_raises(ValueError, odeint, sys1, [0, 1], [10, 10], jacfn=badjac)


if __name__ == "__main__":
    run_module_suite()



# -*- coding: utf-8 -*-
# Authors: B. Malengier based on ode.py
r"""
First-order ODE solver
======================

User-friendly interface to various numerical integrators for solving a
system of first order ODEs with prescribed initial conditions:

.. math::

    \frac{dy(t)}{dt} = f(t, y(t)), \quad y(t_0)=y_0

    y(t_0)[i] = y_0[i], i = 0, ..., \mathrm{len}(y_0) - 1

:math:`f(t,y)` is the right hand side function and returns a vector of size
:math:`\mathrm{len}(y_0)`.

"""

from __future__ import print_function

__all__ = ['ode']
__version__ = "$Id$"
__docformat__ = "restructuredtext en"

import re, sys

from numpy import isscalar, array, asarray

class OdeBase(object):
    """
    The interface which ODE solvers must implement.

    Parameters
    ----------
    Rfn : function
        A function which computes the required derivatives. The signature should be ``func(t, y, y_dot, *args, **kwargs)``. Note that *args and **kwargs handling are solver dependent.

    options : mapping
        Additional options for initialization, solver dependent
    """

    integrator_classes = []

    def __init__(self, Rfn, **options):
        raise NotImplementedError('all ODE solvers must implement this')

    def set_options(self, **options):
        """
        Set specific options for the solver.

        Calling set_options a second time, normally resets the solver.
        """
        raise NotImplementedError('all ODE solvers must implement this')

    def solve(self, tspan, y0):
        """
        Runs the solver.

        Parameters
        ----------
        tspan : array (or similar)
            a list of times at which the computed value will be returned. Must
            contain the start time as first entry.

        y0 : array (or similar)
            a list of initial values

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnum)
            ``values``  Named tuple with entries t and y
            ``errors``  Named tuple with entries t and y
            ``roots``   Named tuple with entries t and y
            ``tstop``   Named tuple with entries t and y
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   indicating return status of the solver
            ``t``      numpy array of times at which the computations were successful
            ``y``      numpy array of values corresponding to times t (values of y[i, :] ~ t[i])
            ``t_err``  float or None - if recoverable error occured (for example reached maximum number of allowed iterations), this is the time at which it happened
            ``y_err``  numpy array of values corresponding to time t_err
            ========== ==========================================
        """
        raise NotImplementedError('all ODE solvers must implement this')

    def init_step(self, t0, y0):
        """
        Initializes the solver and allocates memory.

        Parameters
        ----------
        t0 : number
            initial time
        y0 : array
            initial condition for y (can be list or numpy array)

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnum)
            ``values``  Named tuple with entries t and y
            ``errors``  Named tuple with entries t and y
            ``roots``   Named tuple with entries t and y
            ``tstop``   Named tuple with entries t and y
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   boolean status of the computation (successful or error occured)
            ``t_out``  inititial time
            ========== ==========================================
        """
        raise NotImplementedError('all ODE solvers must implement this')

    def step(self, t, y_retn=None):
        """
        Method for calling successive next step of the ODE solver to allow
        more precise control over the solver. The 'init_step' method has to
        be called before the 'step' method.

        A step is done towards time t, and output at t returned.  This time can be higher or lower than the previous time.  If option 'one_step_compute'==True, and the solver supports it, only one internal solver step is done in the direction of t starting at the current step.

        If old_api=True, the old behavior is used: if t>0.0 then integration is performed until this time and results at this time are returned in y_retn if t<0.0 only one internal step is perfomed towards time abs(t) and results after this one time step are returned

        Parameters
        ----------
        t : number

        y_retn : numpy vector (ndim = 1)
            in which the computed value will be stored  (needs to be preallocated).  If None y_retn is not used.

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnum)
            ``values``  Named tuple with entries t and y
            ``errors``  Named tuple with entries t and y
            ``roots``   Named tuple with entries t and y
            ``tstop``   Named tuple with entries t and y
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   status of the computation (successful or error occured)
            ``t_out``  time, where the solver stopped (when no error occured, t_out == t)
            ========== ==========================================
        """
        raise NotImplementedError('all ODE solvers must implement this')

#------------------------------------------------------------------------------
# User interface
#------------------------------------------------------------------------------

class ode(object):
    """
    A generic interface class to differential equation solvers.


    Parameters
    ----------

    integrator_name : ``'cvode'``, ``'dopri5'`` or ``'dop853'``
        The solver to use.

        ``'cvode'`` selects the CVODE solver from the SUNDIALS package.
        See py:class:`scikits.odes.sundials.cvode.CVODE` for solver specific
        options.

        ``'dopri5'`` selects the Dormand & Prince Runge-Kutta order (4)5
        solver from scipy.

    eqsrhs : right-hand-side function
        Right-hand-side of a first order ode.
        Generally, you can assume the following signature to work:

            eqsrhs(x, y, return_rhs)

        with

            x: independent variable, eg the time, float

            y: array of n unknowns in x

            return_rhs : array that must be updated with the value of the
            right-hand-side, so f(t,y).  The dimension is equal to
            dim(y)

        return value: An integer, 0 for success, 1 for failure.
            It is not guaranteed that a solver takes this status into account

        Some solvers will allow userdata to be passed to eqsrhs, or optional
        formats that are more performant.

    options :  additional options of the solver
        See set_options method of the `integrator_name` you selected for
        details.
        Set option `old_api=False` to use the new API. In the future, this
        will become the default!

    See Also
    --------
    scikits.odes.odeint.odeint : an ODE integrator with a simpler interface
    scipy.integrate : Methods in scipy for ODE integration

    Examples
    --------
    ODE arise in many applications of dynamical systems, as well as in
    discritisations of PDE (eg moving mesh combined with method of
    lines).
    As an easy example, consider the simple oscillator,

    >>> from __future__ import print_function
    >>> from numpy import cos, sin, sqrt
    >>> k = 4.0
    >>> m = 1.0
    >>> initx = [1, 0.1]
    >>> def rhseqn(t, x, xdot):
            # we create rhs equations for the problem
            xdot[0] = x[1]
            xdot[1] = - k/m * x[0]

    >>> from scikits.odes import ode
    >>> solver = ode('cvode', rhseqn, old_api=False)
    >>> result = solver.solve([0., 1., 2.], initx)
    >>> print('   t        Solution          Exact')
    >>> print('------------------------------------')
    >>> for t, u in zip(result.values.t, result.values.y):
            print('%4.2f %15.6g %15.6g' % (t, u[0], initx[0]*cos(sqrt(k/m)*t)+initx[1]*sin(sqrt(k/m)*t)/sqrt(k/m)))

    More examples in the Examples_ directory and IPython_ worksheets.

    .. _Examples: https://github.com/bmcage/odes/tree/master/docs/src/examples
    .. _IPython: https://github.com/bmcage/odes/tree/master/docs/ipython
    """
    LOADED = False

    def __init__(self, integrator_name, eqsrhs, **options):
        integrator = find_ode_integrator(integrator_name)
        if integrator is None:
            raise ValueError('No integrator name match with %s or is not available.'\
                  %(repr(integrator_name)))
        else:
            self._integrator = integrator(eqsrhs, **options)

    def set_options(self, **options):
        """
        Set specific options for the solver.
        See the solver documentation for details.

        Calling set_options a second time, is only possible for options that
        can change during runtime.
        """
        return self._integrator.set_options(**options)

    def solve(self, tspan, y0):
        """
        Runs the solver.

        Parameters
        ----------
        tspan : array (or similar)
            a list of times at which the computed value will be returned. Must
            contain the start time as first entry.

        y0 : array (or similar)
            a list of initial values

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnum)
            ``values``  Named tuple with entries t and y
            ``errors``  Named tuple with entries t and y
            ``roots``   Named tuple with entries t and y
            ``tstop``   Named tuple with entries t and y
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   indicating return status of the solver
            ``t``      numpy array of times at which the computations were successful
            ``y``      numpy array of values corresponding to times t (values of y[i, :] ~ t[i])
            ``t_err``  float or None - if recoverable error occured (for example reached maximum number of allowed iterations), this is the time at which it happened
            ``y_err``  numpy array of values corresponding to time t_err
            ========== ==========================================

        """
        return self._integrator.solve(tspan, y0)

    def init_step(self, t0, y0):
        """
        Initializes the solver and allocates memory. It is not needed to
        call this method if solve is used to compute the solution. In the case
        step is used, init_step must be called first.

        Parameters
        ----------
        t0 : number
            initial time
        y0 : array
            initial condition for y (can be list or numpy array)

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnum)
            ``values``  Named tuple with entries t and y
            ``errors``  Named tuple with entries t and y
            ``roots``   Named tuple with entries t and y
            ``tstop``   Named tuple with entries t and y
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   boolean status of the computation (successful or error occured)
            ``t_out``  inititial time
            ========== ==========================================

        """
        return self._integrator.init_step(t0, y0)

    def step(self, t, y_retn=None):
        """
        Method for calling successive next step of the ODE solver to allow
        more precise control over the solver. The 'init_step' method has to
        be called before the 'step' method.

        A step is done towards time t, and output at t returned.  This time can be higher or lower than the previous time.  If option 'one_step_compute'==True, and the solver supports it, only one internal solver step is done in the direction of t starting at the current step.

        If old_api=True, the old behavior is used: if t>0.0 then integration is performed until this time and results at this time are returned in y_retn if t<0.0 only one internal step is perfomed towards time abs(t) and results after this one time step are returned

        Parameters
        ----------
        t : number

        y_retn : numpy vector (ndim = 1)
            in which the computed value will be stored  (needs to be preallocated).  If None y_retn is not used.

        Returns
        -------
        old_api is False : namedtuple
            namedtuple with the following attributes

            =========== ==========================================
            Field       Meaning
            =========== ==========================================
            ``flag``    An integer flag (StatusEnum)
            ``values``  Named tuple with entries t and y
            ``errors``  Named tuple with entries t and y
            ``roots``   Named tuple with entries t and y
            ``tstop``   Named tuple with entries t and y
            ``message`` String with message in case of an error
            =========== ==========================================

        old_api is True : tuple
            tuple with the following elements in order

            ========== ==========================================
            Field      Meaning
            ========== ==========================================
            ``flag``   status of the computation (successful or error occured)
            ``t_out``  time, where the solver stopped (when no error occured, t_out == t)
            ========== ==========================================

        """
        return self._integrator.step(t, y_retn)

    def set_tstop(self, tstop):
        """
        Add a stop time to the integrator past which he is not allowed to
        integrate.

        Parameters
        ----------
        tstop : float time
            Time point in the future where the integration must stop. You can
            indicate like this that integration past this point is not allowed,
            in order to avoid undefined behavior.
            You can achieve the same result with a call to
            `set_options(tstop=tstop)`
        """
        if hasattr(self._integrator, 'set_tstop'):
            self._integrator.set_tstop(tcrit)
        else:
            self._integrator.set_options(tstop=tstop)

    def get_info(self):
        """
        Return additional information about the state of the integrator.

        Returns
        -------

        A dictionary filled with internal data as exposed by the integrator.
        See the `get_info` method of your chosen integrator for details.

        """
        if hasattr(self._integrator, 'get_info'):
            return self._integrator.get_info()
        else:
            return {}

#------------------------------------------------------------------------------
# ODE integrators
#------------------------------------------------------------------------------

def find_ode_integrator(name):
    if not ode.LOADED:
        ## cvode
        try:
            from .sundials import cvode
            OdeBase.integrator_classes.append(cvode.CVODE)
        except ValueError as msg:
            print('Could not load CVODE solver', msg)
        except ImportError:
            print(sys.exc_info()[1])
        ## cvodes
        try:
            from .sundials import cvodes
            OdeBase.integrator_classes.append(cvodes.CVODES)
        except ValueError as msg:
            print('Could not load CVODES solver', msg)
        except ImportError:
            print(sys.exc_info()[1])

        ## dopri5 and dop853
        try:
            from .dopri5 import dopri5, dop853
        except ImportError:
            print(sys.exc_info()[1])

        ode.LOADED = True

    for cl in OdeBase.integrator_classes:
        if re.match(name, cl.__name__, re.I):
            return cl
        elif hasattr(cl, name) and re.match(name, cl.name, re.I):
            return cl
    raise ValueError('Integrator name %s does not exist' % name)



# -*- coding: utf-8 -*-
"""
Created on Fri Jan 29 12:58:13 2016

@author: benny
"""

from __future__ import division, print_function, absolute_import

import numpy as np

from numpy.testing import (
    assert_, TestCase, run_module_suite, assert_array_almost_equal,
    assert_raises, assert_allclose, assert_array_equal, assert_equal)
from scikits.odes import ode
from scikits.odes.dopri5 import StatusEnumDOP


class SimpleOscillator():
    r"""
    Free vibration of a simple oscillator::
        m \ddot{u} + k u = 0, u(0) = u_0 \dot{u}(0) \dot{u}_0
    Solution::
        u(t) = u_0*cos(sqrt(k/m)*t)+\dot{u}_0*sin(sqrt(k/m)*t)/sqrt(k/m)
    """
    stop_t = 1 + 0.09
    y0 = np.array([1.0, 0.1], float)

    k = 4.0
    m = 1.0
    atol = 1e-6
    rtol = 1e-6

    def f(self, t, y, out):
        tmp = np.zeros((2, 2), float)
        tmp[0, 1] = 1.0
        tmp[1, 0] = -self.k / self.m
        out[:] = np.dot(tmp, y)[:]

    def verify(self, t, y):
        omega = np.sqrt(self.k / self.m)
        u = self.y0[0]*np.cos(omega*t) + self.y0[1]*np.sin(omega*t)/omega
        return np.allclose(u, y[0], atol=self.atol, rtol=self.rtol)

class TestDop(TestCase):

    def test_dop_oscil(self):
        #test calling sequence. End is reached before root is found
        prob = SimpleOscillator()
        tspan = [0, prob.stop_t]
        solver = ode('dopri5', prob.f)
        soln = solver.solve(tspan, prob.y0)
        assert soln.flag==StatusEnumDOP.SUCCESS, "ERROR: Error occurred"
        assert prob.verify(prob.stop_t, soln.values.y[1])

        solver = ode('dop853', prob.f)
        soln = solver.solve(tspan, prob.y0)
        assert soln.flag==StatusEnumDOP.SUCCESS, "ERROR: Error occurred"
        assert prob.verify(prob.stop_t, soln.values.y[1])



# -*- coding: utf-8 -*-
# Created on Thu Jan 28 14:59:25 2016
# @author: benny
"""
Making scipy ode solvers available via the ode API
==================================================

dopri5
------
    This is an explicit runge-kutta method of order (4)5 due to Dormand & Prince
    (with stepsize control and dense output).
    The API of this solver is as the other scikit.odes ODE solvers

    Authors:

        E. Hairer and G. Wanner Universite de Geneve, Dept. de Mathematiques CH-1211 Geneve 24, Switzerland e-mail: ernst.hairer@math.unige.ch, gerhard.wanner@math.unige.ch

    This code is described in [HNW93].

    This integrator accepts the following options:

        atol : float or sequence absolute tolerance for solution, default 1e-12
        rtol : float or sequence relative tolerance for solution, default 1e-6
        nsteps : int Maximum number of (internally defined) steps allowed during one call to the solver. Default=500
        first_step : float
        max_step : float
        safety : float Safety factor on new step selection (default 0.9)
        ifactor : float
        dfactor : float Maximum factor to increase/decrease step size by in one step
        beta : float Beta parameter for stabilised step size control.
        verbosity : int Switch for printing messages (< 0 for no messages).

References
[HNW93]	(1, 2) E. Hairer, S.P. Norsett and G. Wanner, Solving Ordinary Differential Equations i. Nonstiff Problems. 2nd edition. Springer Series in Computational Mathematics, Springer-Verlag (1993)

dop853
------
    This is an explicit runge-kutta method of order 8(5,3) due to Dormand & Prince
    (with stepsize control and dense output).
    Options and references the same as “dopri5”.

"""

from __future__ import print_function

from collections import namedtuple
from enum import IntEnum
from warnings import warn

import numpy as np

from .ode import OdeBase


SolverReturn = namedtuple(
    "SolverReturn", [
        "flag", "values", "errors", "roots",
        "tstop", "message"
    ]
)

SolverVariables = namedtuple("SolverVariables", ["t", "y"])

class StatusEnumDOP(IntEnum):
    SUCCESS = 1
    SOLOUT  = 2

    INPUT_FAIL     = -1
    NMAX_FAIL      = -2
    STEPSIZE_FAIL  = -3
    STIFFNESS_FAIL = -4

    UNEXPECTED_IDID = -10

STATUS_MESSAGE = {
    StatusEnumDOP.SUCCESS: 'computation successful',
    StatusEnumDOP.SOLOUT: 'comput. successful (interrupted by solout)',
    StatusEnumDOP.INPUT_FAIL: 'input is not consistent',
    StatusEnumDOP.NMAX_FAIL: 'larger nmax is needed',
    StatusEnumDOP.STEPSIZE_FAIL: 'step size becomes too small',
    StatusEnumDOP.STIFFNESS_FAIL: 'problem is probably stiff (interrupted)',
    StatusEnumDOP.UNEXPECTED_IDID: 'Unexpected idid, check warnings for info',
}

class DOPSolveException(Exception):
    """Base class for exceptions raised by `DOP.validate_flags`."""
    def __init__(self, soln):
        self.soln = soln
        self.args = (self._message.format(soln),)

class DOPSolveFailed(DOPSolveException):
    """`DOP.solve` failed to reach endpoint"""
    _message = (
        "Solver failed with flag {0.flag} and finished at {0.errors.t}"
        "with values {0.errors.y}."
    )

class dopri5(OdeBase):
    try:
        from scipy.integrate import ode as _runner
    except ImportError:
        print(sys.exc_info()[1])
        _runner = None

    name = 'dopri5'

    default_values = {
            'rtol': 1e-6,
            'atol': 1e-12,
            'nsteps': 500,
            'max_step': 0.0,
            'first_step': 0.0,  # determined by solver
            'safety': 0.9,
            'ifactor': 10.0,
            'dfactor': 0.2,
            'beta': 0.0,
            'verbosity': -1, # no messages if negative
            }

    def __init__(self, Rfn, **options):
        """
        Initialize the ODE Solver and it's default values

        Parameters
        ----------
            Rfn     - right-hand-side function
            options - additional options for initialization
        """
        self.options = self.default_values
        self.set_options(rfn=Rfn, **options)
        self._validate_flags = None
        self.solver = None
        self.initialized = False

    def set_options(self, **options):
        """
        Set specific options for the solver.

        Calling set_options a second time, normally resets the solver.
        """
        for (key, value) in options.items():
            self.options[key.lower()] = value
        self.initialized = False

    def _init_data(self):
        self.rfn = self.options['rfn']
        self.N = len(self.y0)

    def _wrap_Rfn(self, t, y):
        yout = np.empty(self.N, float)
        self.rfn(t, y, yout)
        return yout

    def init_step(self, t0, y0):
        """
        Initializes the solver and allocates memory.

        Parameters
        ----------
            t0     - initial time
            y0     - initial condition for y (can be list or numpy array)

        Returns
        -------
         if old_api:
            not supported

         if old_api False:
            A named tuple, with entries:
                flag   = An integer flag (StatusEnumDop)
                values = Named tuple with entries t and y and ydot. y will
                            correspond to y_retn value and ydot to yp_retn!
                errors = Named tuple with entries t_err and y_err
                roots  = Named tuple with entries t_roots and y_roots
                tstop  = Named tuple with entries t_stop and y_tstop
                message= String with message in case of an error

        """
        self.y0 = y0
        self.t = t0
        self._init_data()

        self.solver = self._runner(self._wrap_Rfn)\
                        .set_integrator(self.name,
                            rtol = self.options['rtol'],
                            atol = self.options['atol'],
                            nsteps = self.options['nsteps'],
                            max_step = self.options['max_step'],
                            first_step = self.options['first_step'],  # determined by solver
                            safety = self.options['safety'],
                            ifactor = self.options['ifactor'],
                            dfactor = self.options['dfactor'],
                            beta = self.options['beta'],
                            verbosity = self.options['verbosity'])\
                        .set_initial_value(y0, t0)
        self.initialized = True
        y_retn  = np.empty(np.alen(y0), float)
        y_retn[:] = y0[:]
        soln = SolverReturn(
            flag=StatusEnumDOP.SUCCESS,
            values=SolverVariables(t=t0, y=y_retn),
            errors=SolverVariables(t=None, y=None),
            roots=SolverVariables(t=None, y=None),
            tstop=SolverVariables(t=None, y=None),
            message=STATUS_MESSAGE[StatusEnumDOP.SUCCESS]
        )
        if self._validate_flags:
            return self.validate_flags(soln)
        return soln

    def step(self, t, y_retn=None):
        """
        Method for calling successive next step of the ODE solver to allow
        more precise control over the solver. The 'init_step' method has to
        be called before the 'step' method.

        Parameters
        ----------
            t - A step is done towards time t, and output at t returned.
                This time can be higher or lower than the previous time.
                If option 'one_step_compute'==True, and the solver supports
                it, only one internal solver step is done in the direction
                of t starting at the current step.

                If old_api=True, the old behavior is used:
                 if t>0.0 then integration is performed until this time
                          and results at this time are returned in y_retn
                 if t<0.0 only one internal step is perfomed towards time abs(t)
                         and results after this one time step are returned
            y_retn - numpy vector (ndim = 1) in which the computed
                     value will be stored  (needs to be preallocated).  If
                     None y_retn is not used.
        Returns
        -------
         if old_api:
             not supported

         if old_api False:
            A named tuple, with entries:
                flag   = An integer flag (StatusEnumDOP)
                values = Named tuple with entries t and y. y will
                            correspond to y_retn value
                errors = Named tuple with entries t_err and y_err
                roots  = Named tuple with entries t_roots and y_roots
                tstop  = Named tuple with entries t_stop and y_tstop
                message= String with message in case of an error

        """
        if not self.initialized == False:
            raise ValueError("DOPRI:step: init_step must be run before running the step method.")
        if t <= self.t:
            raise ValueError("Integration must be forward! t must be > previous timestep")
        y = self.solver.integrate(t)
        y_err = None
        t_err = None
        if not self.solver.successful():
            flag = StatusEnumDOP.UNEXPECTED_IDID
            y_err = y
            t_err = t
        else:
            flag = StatusEnumDOP.SUCCESS
            self.y = y
            self.t = self.solver.t

        return SolverReturn(
            flag=flag,
            values=SolverVariables(t=self.t, y=y),
            errors=SolverVariables(t=t_err, y=y_err),
            roots=SolverVariables(t=None, y=None),
            tstop=SolverVariables(t=None, y=None),
            message=STATUS_MESSAGE[flag]
        )

    def solve(self, tspan, y0):
        """
        Runs the solver.

        Parameters
        ----------
            tspan - an list/array of times at which the computed value will be
                    returned. Must contain the start time as first entry..
            y0    - list/numpy array of initial values

        Returns
        -------
         if old_api
            Not supported
         if old_api False:
            A named tuple, with entries:
                flag   = An integer flag
                values = Named tuple with entries t and y
                errors = Named tuple with entries t and y
                roots  = Named tuple with entries t and y
                tstop  = Named tuple with entries t and y
                message= String with message in case of an error
        """
        self.initialized = False
        soln = self.init_step(tspan[0], y0)
        nrt = len(tspan)
        t_retn  = np.empty(np.shape(tspan), float)
        y_retn  = np.empty([np.alen(tspan), np.alen(y0)], float)

        t_retn[0] = tspan[0]
        y_retn[0, :] = y0

        idx = 1
        while self.solver.successful() and idx<nrt:
            time = tspan[idx]
            y = self.solver.integrate(time)
            t_retn[idx] = time
            y_retn[idx, :] = y
            idx += 1

        y_err = None
        t_err = None
        if not self.solver.successful():
            flag = StatusEnumDOP.UNEXPECTED_IDID
            y_err = y_retn[idx-1]
            t_err = t_retn[idx-1]
            # return values computed so far
            t_retn  = t_retn[0:idx-1]
            y_retn  = y_retn[0:idx-1, :]
        else:
            flag = StatusEnumDOP.SUCCESS

        soln = SolverReturn(
            flag=flag,
            values=SolverVariables(t=t_retn, y=y_retn),
            errors=SolverVariables(t=t_err, y=y_err),
            roots=SolverVariables(t=None, y=None),
            tstop=SolverVariables(t=None, y=None),
            message=STATUS_MESSAGE[flag]
        )
        if self._validate_flags:
            return self.validate_flags(soln)
        return soln

    def validate_flags(self, soln):
        """
        Validates the flag returned by `dopri.solve`.

        Validation happens using the following scheme:
         * failures (`flag` < 0) raise `DOPSolveFailed` or a subclass of it;
         * otherwise, return an instance of `SolverReturn`.

        """
        if soln.flag == StatusEnumDOP.SUCCESS:
            return soln
        if soln.flag < 0:
            raise DOPSolveFailed(soln)
        warn(WARNING_STR.format(soln.flag, *soln.err_values))
        return soln

class dop853(dopri5):
    name = 'dop853'
    default_values = {
            'rtol': 1e-6,
            'atol': 1e-12,
            'nsteps': 500,
            'max_step': 0.0,
            'first_step': 0.0,  # determined by solver
            'safety': 0.9,
            'ifactor': 6.0,
            'dfactor': 0.3,
            'beta': 0.0,
            'verbosity': -1, # no messages if negative
            }


if dopri5._runner:
    OdeBase.integrator_classes.append(dopri5)

if dop853._runner:
    OdeBase.integrator_classes.append(dop853)



#
# odes - Extra ode integrators
#

import inspect

class CVODESolveException(Exception):
    """Base class for exceptions raised by ``CVODE.validate_flags``."""
    def __init__(self, soln):
        self.soln = soln
        self.args = (self._message.format(soln),)

class CVODESolveFailed(CVODESolveException):
    """``CVODE.solve`` failed to reach endpoint"""
    _message = (
        "Solver failed with flag {0.flag} and finished at {0.errors.t}"
        "with values {0.errors.y}."
    )

class CVODESolveFoundRoot(CVODESolveException):
    """``CVODE.solve`` found a root"""
    _message = "Solver found a root at {0.roots.t[0]}."

class CVODESolveReachedTSTOP(CVODESolveException):
    """``CVODE.solve`` reached the endpoint specified by tstop."""
    _message = "Solver reached tstop at {0.tstop.t[0]}."

class IDASolveException(Exception):
    """Base class for exceptions raised by ``IDA.validate_flags``."""
    def __init__(self, soln):
        self.soln = soln
        self.args = (self._message.format(soln),)

class IDASolveFailed(IDASolveException):
    """``IDA.solve`` failed to reach endpoint"""
    _message = (
        "Solver failed with flag {0.flag} and finished at {0.errors.t}"
        "with values {0.errors.y} and derivatives {0.errors.ydot}."
    )

class IDASolveFoundRoot(IDASolveException):
    """``IDA.solve`` found a root"""
    _message = "Solver found a root at {0.roots.t[0]}."

class IDASolveReachedTSTOP(IDASolveException):
    """``IDA.solve`` reached the endpoint specified by tstop."""
    _message = "Solver reached tstop at {0.tstop.t[0]}."


def _get_num_args(func):
    """
    Python 2/3 compatible method of getting number of args that `func` accepts
    """
    if hasattr(inspect, "signature"):
        sig = inspect.signature(func)
        numargs = 0
        for param in sig.parameters.values():
            if param.kind in (
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.VAR_POSITIONAL,
            ):
                numargs += 1
        return numargs
    else:
        return len(inspect.getargspec(func).args)



# Copied and modified from the README of https://github.com/bmcage/odes
# BSD-3-Clause, scikit-odes authors

import numpy as np
from scikits.odes import ode


def van_der_pol(t, y, ydot):
    ydot[0] = y[1]
    ydot[1] = 1000*(1.0-y[0]**2)*y[1]-y[0]


t0, y0 = 1, np.array([0.5, 0.5])
ts = np.linspace(t0, 500, 200)
solution = ode('cvode', van_der_pol, old_api=False).solve(ts, y0)
print(solution)



# -*- coding: utf-8 -*-
# Authors: B. Malengier
"""
ddaspk
======
Solver developed 1989 to 1996, with some corrections from 2000 - Fortran

This code solves a system of differential/algebraic equations of the form 
G(t,y,y') = 0 , using a combination of Backward Differentiation Formula 
(BDF) methods and a choice of two linear system solution methods: direct 
(dense or band) or Krylov (iterative). 
Krylov is not supported from within scikits.odes. 
In order to support it, a new interface should be created ddaspk_krylov, 
with a different signature, reflecting the changes needed.

Source: http://www.netlib.org/ode/ddaspk.f

On construction the function calculating the residual (res) must be given and 
optionally also the function calculating the jacobian (jac). 
res has the signature: res(x, y, yprime, return_residual)
with 
    x       : independent variable, eg the time, float
    y       : array of n unknowns in x
    yprime  : dy/dx array of n unknowns in x, dimension = dim(y)
    return_residual: array that must be updated with the value of the residuals,
              so G(t,y,y').  The dimension is equal to dim(y)
    return value: Not needed. However, for use with other solvers, consider 
              returning 0 on success.

jac has the signature jac(x, y, yprime, cj, return_jac)
with 
    x       : independent variable, eg the time, float
    y       : array of n unknowns in x
    yprime  : dy/dx array of n unknowns in x, dimension = dim(y)
    cj      : internal variable of ddaspk algorithm you can use, don't change it! 
              cj can be ignored, or used to rescale constraint equations in the 
              system
    return_jac : two dimensional array of the Jacobian, as per the Jacobian 
              definition of the ddaspk solver. This means it should be or a 
                full nxn shaped array in general (n=dim(y)), or a banded shaped
                array as per the definition of lband/uband. 
                Jac is optional and should be set with the jacfn option keyword. 
                Note that jac is defined as 
                            dres(i)/dy(j) + cj*dres(i)/dyprime(j)
    return value: Not needed. However, for use with other solvers, consider 
              returning 0 on success.

This integrator accepts the following parameters in the initializer or 
set_options method of the dae class:

- rfn : residual function, see above for signature. This option need not be
        set, as rfn will be set during initialization. If the residual function
        of initialization needs to be reset, this option can be used
- jacfn : jacobian function, see above for signature. Default is None.
- atol : float or sequence of length i
  absolute tolerance for solution
- rtol : float or sequence of length i
  relative tolerance for solution
- lband : None or int
- uband : None or int
  Jacobian band width, jac[i,j] != 0 for i-lband <= j <= i+uband.
  Setting these requires your jac routine to return the jacobian
  in packed format, jac_packed[i-j+lband, j] = jac[i,j].
- tstop : None or float. If given, tstop is a critical time point
  beyond which no integration occurs
- max_steps : int
  Maximum number of (internally defined) steps allowed during one
  call to the solver.
- first_step : float
  Set initial stepsize. DAE solver can suffer on the first step, set this
  to circumvent this.
- max_step_size : float
  Limits for the step sizes used by the integrator. This overrides the
  internal value
- order : int
  Maximum order used by the integrator, >=1,  <= 5 for BDF.
  5 is the default
- enforce_nonnegativity: bool
  Enforce the nonnegativity of Y during integration
  Note: best is to run code first with no condition
- compute_initcond: None or 'yp0' or 'y0'
  DDASPK may be able to compute the initial conditions if you do not know them
  precisely. 
  If y0, then y0 will be calculated
  If yp0, then the differential variables will be used to solve for the 
    algebraic variables and the derivative of the differential variables. 
    Which are the algebraic variables must be indicated with algebraic_var method
- exclude_algvar_from_error: bool
  To determine solution, do not take the algebraic variables error control into 
  account. Default=False
- constraint_init: bool
  Enforce the constraint checks of Y during initial condition computation
  Note: try first with no constraints
- constraint_type: if constraint_init, give an integer array with for every
  unknown the specific condition to check: 
       1: y0[i] >= 0 
       2: y0[i] >  0
      -1: y0[i] <= 0
      -2: y0[i] <  0
       0: y0[i] not constrained
  Alternatively, pass only one integer that then applies to all unknowns
- algebraic_vars_idx: an array or None (= default)
                Description:
                    If given problem is of type DAE, some items of the residual
                    vector returned by the 'resfn' have to be treated as 
                    algebraic variables. These are denoted by the position 
                    (index) in the residual vector.
                    All these indexes have to be specified in the 
                    'algebraic_vars_idx' array.
"""

from __future__ import print_function

__all__ = ['ddaspk']
__version__ = "$Id$"
__docformat__ = "restructuredtext en"

from numpy import asarray, array, zeros, sin, int32, isscalar, empty, alen
from copy import copy
from .dae import DaeBase
import re, sys

class ddaspk(DaeBase):
    try:
        from .ddaspk import ddaspk as _runner
    except ImportError:
        print(sys.exc_info()[1])
        _runner = None
    # _runner = getattr(_ddaspk,'ddaspk',None)
    name = 'ddaspk'

    messages = { 1: 'A step was successfully taken in the '
                    'intermediate-output mode.  The code has not '
                    'yet reached TOUT.', 
                 2: 'The integration to TSTOP was successfully '
                    'completed (T = TSTOP) by stepping exactly to TSTOP.', 
                 3: 'The integration to TOUT was successfully '
                    'completed (T = TOUT) by stepping past TOUT. '
                    'Y(*) and YPRIME(*) are obtained by interpolation.', 
                 4: 'The initial condition calculation, with '
                    'INFO(11) > 0, was successful, and INFO(14) = 1. '
                    'No integration steps were taken, and the solution '
                    'is not considered to have been started.', 
                -1: 'A large amount of work has been expended (about 500 steps)',
                -2: 'Excess accuracy requested. (Tolerances too small.)',
                -3: 'The local error test cannot be satisfied because you '
                    'specified a zero component in ATOL and the corresponding'
                    ' computed solution component is zero.  Thus, a pure'
                    ' relative error test is impossible for this component.',
                -5: 'Repeated failures in the evaluation or processing of the'
                    ' preconditioner (in JAC)',
                -6: 'repeated error test failures on the last attempted step)', 
                -7: 'The nonlinear system solver in the time integration could'
                    ' not converge.',
                -8: 'The matrix of partial derivatives appears to be singular'
                    ' (direct method).', 
                -9: 'The nonlinear system solver in the time integration'
                    'failed to achieve convergence, and there were repeated '
                    'error test failures in this step.', 
                -10:'The nonlinear system solver in the time integration failed'
                    ' to achieve convergence because IRES was equal to -1.', 
                -11:'IRES = -2 was encountered and control is'
                    'being returned to the calling program.', 
                -12:'Failed to compute the initial Y, YPRIME.', 
                -13:"Unrecoverable error encountered inside user's"
                    "PSOL routine, and control is being returned to"
                    "the calling program.", 
                -14:'The Krylov linear system solver could not '
                    'achieve convergence.', 
                -33:'The code has encountered trouble from which'
                   ' it cannot recover.  A message is printed'
                   ' explaining the trouble and control is returned'
                   ' to the calling program.  For example, this occurs'
                   ' when invalid input is detected.', 
                }
    supports_run_relax = 0
    supports_step = 1

    def __init__(self, resfn, **options):
        default_values = {
            'rtol': 1e-6,
            'atol': 1e-12,
            'lband': None,
            'uband': None,
            'tstop': None, 
            'order' : 5,
            'max_steps' : 500,
            'max_step_size' : 0.0, # corresponds to infinite
            'first_step' : 0.0, # determined by solver
            'enforce_nonnegativity': False, 
            'nonneg_type': None, 
            'compute_initcond': None,
            'constraint_init': False, 
            'constraint_type': None, 
            'algebraic_vars_idx': None,
            'exclude_algvar_from_error': False, 
            'rfn': None,
            'jacfn': None,
            }
        self.t = None
        self.y = None
        self.yp = None
        self.tmp_res = None
        self.tmp_jac = None
        self.options = default_values
        self.set_options(rfn=resfn, **options)
        self.initialized = False

    def set_options(self, **options):
        """ See dae.DaeBase
        """
        for (key, value) in options.items():
            self.options[key.lower()] = value
        self.initialized = False
        
    def _init_data(self):
        self.rtol = self.options['rtol']
        self.atol = self.options['atol']
        self.mu = self.options['uband']
        self.ml = self.options['lband']
        self.jac = self.options['jacfn']
        self.res = self.options['rfn']

        self.tstop = self.options['tstop']
        if self.options['order'] > 5 or self.options['order'] < 1:
            raise ValueError('order should be >=1, <=5')
        self.order = self.options['order']
        self.nsteps = self.options['max_steps']
        self.max_step = self.options['max_step_size']
        self.first_step = self.options['first_step']
        self.nonneg =0 
        if self.options['enforce_nonnegativity'] and self.options['constraint_init']: 
            self.nonneg = 3
        elif self.options['enforce_nonnegativity']: 
            self.nonneg = 2
        elif self.options['constraint_init']: 
            self.nonneg = 1
        if (self.nonneg == 1 or self.nonneg == 3) and self.options['constraint_type'] is None:
            raise ValueError('Give type of init cond contraint as '\
                              'an int array (>=0, >0, <=0, <0) or as int')
        else: self.constraint_type = self.options['constraint_type']
        if self.options['compute_initcond'] is None: 
            self.compute_initcond = 0
        elif re.match(self.options['compute_initcond'], r'y0', re.I): 
            self.compute_initcond = 2
        elif re.match(self.options['compute_initcond'], r'yp0', re.I): 
            self.compute_initcond = 1
        else: 
            raise ValueError('Unknown init cond calculation method %s' %(
                                            self.options['compute_initcond']))
        if self.compute_initcond == 1 and not self.options['algebraic_vars_idx']:
            raise ValueError('Give array indicating which are the '\
                    'algebraic variables with the algebraic_vars_idx option')
        self.algvaridx = self.options['algebraic_vars_idx']
        self.excl_algvar_err = self.options['exclude_algvar_from_error']

        self.success = 1
    
    def init_step(self, t0, y0, yp0, y_ic0_retn = None, yp_ic0_retn = None):
        """ See dae.DaeBase
        """
        self._init_data()
        self.y0 = y0
        self.yp0 = yp0
        self.t = t0
        self._reset(len(y0), self.options['jacfn'] is not None)
        
        if self.compute_initcond:
            if self.first_step <= 0.:
                first_step = 1e-18
            else:
                first_step = self.first_step
            self.y, self.yp, t = self.__run([3, 4], y0, yp0, t0, t0 + first_step)
            t0_init = t
        else:
            self.y = copy(y0)
            self.yp = copy(yp0)
            t0_init = t0
        if not y_ic0_retn is None: y_ic0_retn[:] = self.y[:]
        if not yp_ic0_retn is None: yp_ic0_retn[:] = self.yp[:]

        self.initialized = True
        return t0_init

    def _reset(self, n, has_jac):
        # Calculate parameters for Fortran subroutine ddaspk.
        self.neq = n
        self.algebraic_var = [1]*self.neq
        if self.algvaridx is not None:
            for ind in self.algvaridx:
                self.algebraic_var[ind] = -1
        self.info = zeros((20,), int32)  # default is all info=0
        self.info[17] = 2  # extra output on init cond computation
        if (isscalar(self.atol) != isscalar(self.rtol)) or (
               not isscalar(self.atol) and len(self.atol) != len(self.rtol)):
            raise ValueError('atol (%s) and rtol (%s) must be both scalar or'\
                    ' both arrays of length %s' % (self.atol, self.rtol, n))
        if not isscalar(self.atol):
            self.info[1] = 1
        if has_jac:
            self.info[4] = 1
        if self.mu is not None or self.ml is not None:
            if self.mu is None: self.mu = 0
            if self.ml is None: self.ml = 0
            self.info[5] = 1
        if self.excl_algvar_err:
            self.info[15] = 1
        lrw = 50 + max(self.order+4,7)*n
        if self.info[5]==0: lrw += pow(n, 2)
        elif self.info[4]==0: 
            lrw += (2*self.ml+self.mu+1)*n + 2*(n/(self.ml+self.mu+1)+1)
        else: lrw += (2*self.ml+self.mu+1)*n
        if self.info[15] == 1: lrw +=n
        rwork = zeros((lrw,), float)
        liw = 40 + n
        if self.nonneg in [1, 3]: liw += n
        if self.compute_initcond or self.excl_algvar_err: liw += n
        iwork = zeros((liw,), int32)
        if self.tstop is not None:
            self.info[3] = 1
            rwork[0] = self.tstop
        if self.max_step > 0.0 :
            self.info[6] = 1
            rwork[1] = self.max_step
        if self.first_step > 0.0 :
            self.info[7] = 1
            rwork[2] = self.first_step
        self.rwork = rwork
        if self.ml is not None:
            iwork[0] = self.ml
        if self.mu is not None:
            iwork[1] = self.mu
        if self.order < 5 :
            self.info[8] = 1
            iwork[2] = self.order
        iwork[5] = self.nsteps
        iwork[6] = 2           # mxhnil
        self.info[9] = self.nonneg
        lid = 40
        if self.info[9]==1 or self.info[9]==3 :
            lid = 40 + n
            if isscalar(self.constraint_type): 
                iwork[40:lid]=self.constraint_type
            else: iwork[40:lid]=self.constraint_type[:]
        self.info[10]=self.compute_initcond
        if self.info[10] in [1, 2]:
            iwork[lid:lid+n] = self.algebraic_var[:]
        if self.excl_algvar_err:
            iwork[lid:lid+n] = self.algebraic_var[:]
        ## some overrides that one might want
        # self.info[17] = 1  # minimal printing inside init cond calc
        # self.info[17] = 2  # full printing inside init cond calc
        if self.tstop is not None:
            self.info[3] = 1
            self.rwork[0] = self.tstop
        else:
            self.info[3] = 0
            self.rwork[0] = 0.

        self.iwork = iwork
        
        self.call_args = [self.info,self.rtol,self.atol,self.rwork,self.iwork]        #create storage
        self.tmp_res = empty(self.neq, float)
        if (self.ml is None or self.mu is None) :
            self.tmp_jac = zeros((self.neq, self.neq), float)
        else:
            self.tmp_jac = zeros((self.ml + self.mu + 1, self.neq), float)

        self.success = 1

    def _resFn(self, t, y, yp):
        """Wrapper around the residual as defined by user, to create
         the residual needed by ddaspk
        """
        self.res(t, y, yp, self.tmp_res)
        return self.tmp_res

    def _jacFn(self, t, y, yp, cj):
        """Wrapper around the jacobian as defined by user, to create
         the jacobian needed by ddaspk
        """
        self.jac(t, y, yp, cj, self.tmp_jac)
        return self.tmp_jac

    def solve(self, tspan, y0,  yp0, hook_fn = None):
        """ See dae.DaeBase
        """
        
        t_retn  = empty([alen(tspan), ], float)
        y_retn  = empty([alen(tspan), alen(y0)], float)
        yp_retn = empty([alen(tspan), alen(y0)], float)
        
        y_ic0_retn  = empty(alen(y0), float)
        yp_ic0_retn  = empty(alen(y0), float)
        tinit = self.init_step(tspan[0], y0, yp0, y_ic0_retn, yp_ic0_retn)
        
        t_retn[0] = tinit
        y_retn[0,:] = y0[:]
        yp_retn[0, :] = yp0[:]
        indbreak = None
        for ind, time in enumerate(tspan[1:]):
            if not self.success:
                indbreak = ind + 1
                break
            result = self.__run([2, 3], y_retn[ind], yp_retn[ind], t_retn[ind], time)
            t_retn[ind+1] = result[2]
            y_retn[ind+1][:] = result[0][:]
            yp_retn[ind+1][:] = result[1][:]
        self.t = t_retn[-1]
        if indbreak is not None:
            self.t = t_retn[indbreak-1]
            return self.success, t_retn[:indbreak], y_retn[:indbreak],\
                   yp_retn[:indbreak], t_retn[indbreak], y_retn[indbreak],\
                   yp_retn[indbreak]
        return self.success, t_retn, y_retn, yp_retn, None, None, None

    def __run(self, states, *args):
        # args are: y0,yprime0,t0,t1,res_params,jac_params
        y1, y1prime, t, self.flag = self._runner(*( (self._resFn, self._jacFn) \
                                           + args[:4] + tuple(self.call_args)))
        if self.flag < 0:
            print('ddaspk:',self.messages.get(self.flag,
                                        'Unexpected istate=%s' % self.flag))
            self.success = 0
        elif self.flag not in states:
            print('ddaspk: Run successfull. Unexpected istate=%s, stopping' %
                                            self.flag)
            print(self.messages.get(self.flag, 'Unknown istate=%s' % self.flag))
            self.success = 0
        return y1, y1prime, t

    def step(self, t, y_retn, yp_retn = None):
        """ See dae.DaeBase
        """
        if not self.initialized:
            raise ValueError('Method ''init_step'' has to be called prior to the'
                    'first call of ''step'' method, or after changing options')

        if t > 0.0:
            self.y, self.yp, self.t = self.__run([2, 3], self.y, self.yp, self.t, t)
        else:
            self.info[2] = 1
            self.y, self.yp, self.t = self.__run([1, 2], self.y, self.yp, self.t, -t)
            self.info[2] = 0
        y_retn[:] = self.y[:]
        if yp_retn is not None:
            yp_retn[:] = self.yp[:]
        
        return self.flag, self.t

if ddaspk._runner:
    DaeBase.integrator_classes.append(ddaspk)




from numpy.testing import TestCase, run_module_suite

from .. import ode
from ..sundials.cvode import StatusEnum

def normal_rhs(t, y, ydot):
    ydot[0] = t

def complex_rhs(t, y, ydot):
    ydot[0] = t - y

def rhs_with_return(t, y, ydot):
    ydot[0] = t
    return 0

def rhs_problem_late(t, y, ydot):
    ydot[0] = t
    if t > 0.5:
        return 1

def rhs_problem_immediate(t, y, ydot):
    return 1

def rhs_error_late(t, y, ydot):
    ydot[0] = t
    if t > 0.5:
        return -1

def rhs_error_immediate(t, y, ydot):
    return -1

def normal_root(t, y, g):
    g[0] = 1

def root_with_return(t, y, g):
    g[0] = 1
    return 0

def root_late(t, y, g):
    g[0] = 1
    if t > 0.5:
        g[0] = 0

def root_immediate(t, y, g):
    g[0] = 0

def root_error_late(t, y, g):
    g[0] = t
    if t > 0.5:
        return 1

def root_error_immediate(t, y, g):
    return 1

def normal_jac(t, y, fy, J):
    J[0][0] = 0

def jac_with_return(t, y, fy, J):
    J[0][0] = 0
    return 0

def jac_problem_late(t, y, fy, J):
    J[0][0] = 1
    if t > 0:
        return 1

def jac_problem_immediate(t, y, fy, J):
    return 1

def jac_error_late(t, y, fy, J):
    J[0][0] = 1
    if t > 0:
        return -1

def jac_error_immediate(t, y, fy, J):
    return -1


def normal_jac_vec(v, Jv, t, y):
    Jv[0] = 0

def jac_vec_with_return(v, Jv, t, y):
    Jv[0] = 0
    return None

def jac_vec_problem_late(v, Jv, t, y):
    Jv[0] = v[0]
    if t > 0:
        return 1

def jac_vec_problem_immediate(v, Jv, t, y):
    return 1

def jac_vec_error_late(v, Jv, t, y):
    Jv[0] = v[0]
    if t > 0:
        return -1

def jac_vec_error_immediate(v, Jv, t, y):
    return -1

class TestCVodeReturn(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestCVodeReturn, self).__init__(*args, **kwargs)
        self.solvername = "cvode"

    def test_normal_rhs(self):
        solver = ode(self.solvername, normal_rhs, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_rhs_with_return(self):
        solver = ode(self.solvername, rhs_with_return, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_rhs_problem_late(self):
        solver = ode(self.solvername, rhs_problem_late, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.TOO_MUCH_WORK,
            soln.flag
        )

    def test_rhs_problem_immediate(self):
        solver = ode(self.solvername, rhs_problem_immediate, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.FIRST_RHSFUNC_ERR,
            soln.flag
        )

    def test_rhs_error_late(self):
        solver = ode(self.solvername, rhs_error_late, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.RHSFUNC_FAIL,
            soln.flag
        )

    def test_rhs_error_immediate(self):
        solver = ode(self.solvername, rhs_error_immediate, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.RHSFUNC_FAIL,
            soln.flag
        )

    def test_normal_root(self):
        solver = ode(self.solvername, normal_rhs, rootfn=normal_root, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_root_with_return(self):
        solver = ode(self.solvername, normal_rhs, rootfn=root_with_return, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_root_late(self):
        solver = ode(self.solvername, normal_rhs, rootfn=root_late, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.ROOT_RETURN,
            soln.flag
        )

    def test_root_immediate(self):
        solver = ode(self.solvername, normal_rhs, rootfn=root_immediate, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_root_error_late(self):
        solver = ode(self.solvername, normal_rhs, rootfn=root_error_late, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.RTFUNC_FAIL,
            soln.flag
        )

    def test_root_error_immediate(self):
        solver = ode(self.solvername, normal_rhs, rootfn=root_error_immediate,
                nr_rootfns=1, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.RTFUNC_FAIL,
            soln.flag
        )

    def test_normal_jac(self):
        solver = ode(self.solvername, normal_rhs, jacfn=normal_jac, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_jac_with_return(self):
        solver = ode(self.solvername, normal_rhs, jacfn=jac_with_return, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_jac_problem_late(self):
        solver = ode(self.solvername, complex_rhs, jacfn=jac_problem_late, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.CONV_FAILURE,
            soln.flag
        )

    def test_jac_problem_immediate(self):
        solver = ode(self.solvername, normal_rhs, jacfn=jac_problem_immediate,
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.CONV_FAILURE,
            soln.flag
        )

    def test_jac_error_late(self):
        solver = ode(self.solvername, complex_rhs, jacfn=jac_error_late, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.LSETUP_FAIL,
            soln.flag
        )

    def test_jac_error_immediate(self):
        solver = ode(self.solvername, normal_rhs, jacfn=jac_error_immediate,
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.LSETUP_FAIL,
            soln.flag
        )


    def test_normal_jac_vec(self):
        solver = ode(self.solvername, normal_rhs, jac_times_vecfn=normal_jac_vec, old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_jac_vec_with_return(self):
        solver = ode(self.solvername, normal_rhs, jac_times_vecfn=jac_vec_with_return, linsolver="spgmr", old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.SUCCESS,
            soln.flag
        )

    def test_jac_vec_problem_late(self):
        solver = ode(self.solvername, complex_rhs, jac_times_vecfn=jac_vec_problem_late, linsolver="spgmr", old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.TOO_MUCH_WORK,
            soln.flag
        )

    def test_jac_vec_problem_immediate(self):
        solver = ode(self.solvername, normal_rhs,
                jac_times_vecfn=jac_vec_problem_immediate, linsolver="spgmr",
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.TOO_MUCH_WORK,
            soln.flag
        )

    def test_jac_vec_error_late(self):
        solver = ode(self.solvername, complex_rhs, jac_times_vecfn=jac_vec_error_late,
                linsolver="spgmr", old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.LSOLVE_FAIL,
            soln.flag
        )

    def test_jac_vec_error_immediate(self):
        solver = ode(self.solvername, normal_rhs, jac_times_vecfn=jac_vec_error_immediate, linsolver="spgmr",
                old_api=False)
        soln = solver.solve([0, 1], [1])
        self.assertEqual(
            StatusEnum.LSOLVE_FAIL,
            soln.flag
        )

class TestCVodesReturn(TestCVodeReturn):
    def __init__(self, *args, **kwargs):
        super(TestCVodesReturn, self).__init__(*args, **kwargs)
        self.solvername = "cvodes"

if __name__ == "__main__":
    try:
        run_module_suite()
    except NameError:
        test = TestCVodeReturn()
        TestCVodeReturn.test_normal_rhs()



from __future__ import print_function
import numpy as np
import unittest
from scikits.odes import ode


xs = np.linspace(1, 10, 10)


def true_y(x):
    """ just a dummy this test module is not about the integration itself """
    return np.power(x, 2)
    #return np.sin(x) / (x + 0.1)


def rhs(x, y, ydot):
    ydot[:] = 2 * x
    #ydot[:] = (np.cos(x) * (x + 0.1) - np.sin(x)) / np.pow((x + 0.1), 2)


class GetInfoTest(unittest.TestCase):
    def setUp(self):
        self.ode = ode('cvode', rhs, old_api=False)
        self.solution = self.ode.solve(xs, np.array([1]))

    def test_we_integrated_correctly(self):
        true_ys = true_y(xs)
        diff = np.abs(self.solution.values.y[:, 0] - true_ys)
        rel_diff = diff / true_ys
        print("True solution:\n", true_ys)
        print("Integrated:\n", self.solution.values.y[:, 0])
        print("Difference:\n", diff)
        print("Relative:\n", rel_diff)
        assert rel_diff.max() < 1e-6

    def test_get_info_is_exposed_on_ode(self):
        self.ode.get_info()

    def test_get_info_returns_dict(self):
        assert isinstance(self.ode.get_info(), dict)

    def test_ode_exposes_num_rhs_evals(self):
        info = self.ode.get_info()
        print("ode.get_info() =\n", info)
        assert 'NumRhsEvals' in info
        assert info['NumRhsEvals'] > 0

class GetInfoTestSpils(unittest.TestCase):
    def setUp(self):
        self.ode = ode('cvode', rhs, linsolver="spgmr", old_api=False)
        self.solution = self.ode.solve(xs, np.array([1]))

    def test_ode_exposes_num_njtimes_evals(self):
        info = self.ode.get_info()
        print("ode.get_info() =\n", info)
        assert 'NumJtimesEvals' in info
        assert info['NumJtimesEvals'] > 0



# -*- coding: utf-8 -*-
"""
scikits.odes is a scikit offering a cython wrapper around some extra ode/dae
solvers, so they can mature outside of scipy.

It offers wrappers around the following solvers from `SUNDIALS`_
 * CVODE
 * IDA

It additionally offers wrappers around
 * `ddaspk <http://www.netlib.org/ode/ddaspk.f>` (included)
 * `lsodi <http://www.netlib.org/ode/lsodi.f>` (included)

.. _SUNDIALS: https://computation.llnl.gov/casc/sundials/main.html
"""

from .dae import *
from .ode import *

__all__ = ['test'] + [s for s in dir() if not s.startswith('_')]

try:
    from numpy.testing import Tester
    test = Tester().test
except:
    #testing could not be loaded, old numpy version
    pass




from numpy.testing import TestCase, run_module_suite

from .. import dae
from ..sundials.ida import StatusEnumIDA

def normal_rhs(t, y, ydot, res):
    res[0] = ydot - t

def complex_rhs(t, y, ydot, res):
    res[0] = ydot - t + y

def rhs_with_return(t, y, ydot, res):
    res[0] = ydot - t
    return 0

def rhs_problem_late(t, y, ydot, res):
    res[0] = ydot - t
    if t > 0.5:
        return 1

def rhs_problem_immediate(t, y, ydot, res):
    return 1

def rhs_error_late(t, y, ydot, res):
    res[0] = ydot - t
    if t > 0.5:
        return -1

def rhs_error_immediate(t, y, ydot, res):
    return -1

def normal_root(t, y, y_dot, g):
    g[0] = 1

def root_with_return(t, y, y_dot, g):
    g[0] = 1
    return 0

def root_late(t, y, y_dot, g):
    g[0] = 1
    if t > 0.5:
        g[0] = 0

def root_immediate(t, y, y_dot, g):
    g[0] = 0

def root_error_late(t, y, y_dot, g):
    g[0] = t
    if t > 0.5:
        return 1

def root_error_immediate(t, y, y_dot, g):
    return 1

def normal_jac(t, y, ydot, residual, cj, J):
    J[0][0] = cj

def jac_with_return(t, y, ydot, residual, cj, J):
    J[0][0] = cj
    return 0

def jac_problem_late(t, y, ydot, residual, cj, J):
    J[0][0] = cj + 1
    if t > 0:
        return 1

def jac_problem_immediate(t, y, ydot, residual, cj, J):
    return 1

def jac_error_late(t, y, ydot, residual, cj, J):
    J[0][0] = cj + 1
    if t > 0:
        return -1

def jac_error_immediate(t, y, ydot, residual, cj, J):
    return -1

class TestIdaReturn(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestIdaReturn, self).__init__(*args, **kwargs)
        self.solvername = "ida"

    def test_normal_rhs(self):
        solver = dae(self.solvername, normal_rhs, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.SUCCESS,
            soln.flag
        )

    def test_rhs_with_return(self):
        solver = dae(self.solvername, rhs_with_return, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.SUCCESS,
            soln.flag
        )

    def test_rhs_problem_late(self):
        solver = dae(self.solvername, rhs_problem_late, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.TOO_MUCH_WORK,
            soln.flag
        )

    def test_rhs_problem_immediate(self):
        solver = dae(self.solvername, rhs_problem_immediate, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.REP_RES_ERR,
            soln.flag
        )

    def test_rhs_error_late(self):
        solver = dae(self.solvername, rhs_error_late, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.RES_FAIL,
            soln.flag
        )

    def test_rhs_error_immediate(self):
        solver = dae(self.solvername, rhs_error_immediate, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.RES_FAIL,
            soln.flag
        )

    def test_normal_root(self):
        solver = dae(self.solvername, normal_rhs, rootfn=normal_root, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.SUCCESS,
            soln.flag
        )

    def test_root_with_return(self):
        solver = dae(self.solvername, normal_rhs, rootfn=root_with_return, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.SUCCESS,
            soln.flag
        )

    def test_root_late(self):
        solver = dae(self.solvername, normal_rhs, rootfn=root_late, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.ROOT_RETURN,
            soln.flag
        )

    def test_root_immediate(self):
        solver = dae(self.solvername, normal_rhs, rootfn=root_immediate, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.SUCCESS,
            soln.flag
        )

    def test_root_error_late(self):
        solver = dae(self.solvername, normal_rhs, rootfn=root_error_late, nr_rootfns=1,
                old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.RTFUNC_FAIL,
            soln.flag
        )

    def test_root_error_immediate(self):
        solver = dae(self.solvername, normal_rhs, rootfn=root_error_immediate,
                nr_rootfns=1, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.RTFUNC_FAIL,
            soln.flag
        )

    def test_normal_jac(self):
        solver = dae(self.solvername, normal_rhs, jacfn=normal_jac, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.SUCCESS,
            soln.flag
        )

    def test_jac_with_return(self):
        solver = dae(self.solvername, normal_rhs, jacfn=jac_with_return, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.SUCCESS,
            soln.flag
        )

    def test_jac_problem_late(self):
        solver = dae(self.solvername, complex_rhs, jacfn=jac_problem_late, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.CONV_FAIL,
            soln.flag
        )

    def test_jac_problem_immediate(self):
        solver = dae(self.solvername, normal_rhs, jacfn=jac_problem_immediate,
                old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.CONV_FAIL,
            soln.flag
        )

    def test_jac_error_late(self):
        solver = dae(self.solvername, complex_rhs, jacfn=jac_error_late, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.LSETUP_FAIL,
            soln.flag
        )

    def test_jac_error_immediate(self):
        solver = dae(self.solvername, normal_rhs, jacfn=jac_error_immediate, old_api=False)
        soln = solver.solve([0, 1], [1], [0])
        self.assertEqual(
            StatusEnumIDA.LSETUP_FAIL,
            soln.flag
        )


class TestIdasReturn(TestIdaReturn):
    def __init__(self, *args, **kwargs):
        super(TestIdasReturn, self).__init__(*args, **kwargs)
        self.solvername = "idas"

if __name__ == "__main__":
    try:
        run_module_suite()
    except NameError:
        test = TestIdaReturn()
        test.test_normal_rhs()
